from __future__ import annotations

from typing import Callable

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from matplotlib import axes, figure
from matplotlib.ticker import MaxNLocator

import jijbench as jb
from jijbench.visualization.metrics.baseplot.baseplot import BasePlot


def _get_violations_dict(x: pd.Series) -> dict:
    """Get a dictionary of constraint violations from `pd.Series`.

    This function is intended to be used for `jb.Experiment.table` as example.

    Args:
        x (pd.Series): a Series of benchmark result. Expected to include information on constraint violations.

    Returns:
        dict: a dictionary of constraint violations

    Example:
        The code below get the dictionary of constraint violations for each row in experiment.

        ```python
        import jijbench as jb
        import jijzept as jz
        from jijbench.visualization.metrics.constraintplot.constraintplot import _get_violations_dict

        problem = jb.get_problem("TSP")
        instance_data = jb.get_instance_data("TSP")[0][1]
        multipliers1 = {"onehot_time": 0.003, "onehot_location": 0.003}
        multipliers2 = {"onehot_time": 0.3, "onehot_location": 0.3}

        config_path = "XX"
        sa_sampler = jz.JijSASampler(config=config_path)

        bench = jb.Benchmark(
            params = {
                "model": [problem],
                "feed_dict": [instance_data],
                "multipliers": [multipliers1, multipliers2],
            },
            solver = [sa_sampler.sample_model],
        )
        experiment = bench()
        metrics = experiment.table.apply(_get_violations_dict, axis=1)
        ```
    """
    return x.filter(regex="violations").to_dict()


class ConstraintPlot:
    def __init__(self, result: jb.Experiment) -> None:
        """Visualize the metrics of a benchmark result by boxplot.

        Attributes:
            result (jb.Experiment): a benchmark result.
        """
        self.result = result
        self.baseplot = BasePlot(result)

    def boxplot_violations(
        self,
        figsize: tuple[int | float] | None = None,
        title: str | list[str] | None = None,
        title_fontsize: float | None = None,
        constraint_name_fontsize: float | None = None,
        constraint_name_fontrotation: float | None = None,
        ylabel: str | None = None,
        ylabel_size: float | None = None,
        yticks: list[int | float] | None = None,
        **matplotlib_boxplot_kwargs,
    ) -> tuple[tuple[figure.Figure, axes.Subplot]]:
        """Draw a box and whisker plot of the constraint violations of `result` data using matplotlib.boxplot.

        The arguments are passed to matplotlib functions to change the appearance of the plot.
        matplotlib_boxplot_kwargs are passed to matplotlib.boxplot, and defaults to `{showmeans: True, whis: [0, 100]}`.
            showmeans=True shows mean values in markers.
            the outliers are not considered and whiskers match maxima and minima by whis=[0, 100].
        This method returns a figure and axes, so you can post-process them to change the appearance of the plot.
        See also the example below.

        Args:
            figsize (tuple[int | float] | None): the size of figure. The default uses matplotlib's default value.
            title (str | list[str] | None): the title of figure. The default uses the indices of `result.table`.
            title_fontsize (float | None): the fontsize of the title.The default uses matplotlib's default value.
            constraint_name_fontsize (float | None): the fontsize of the constraint name (i.e. xticklabels). The default uses matplotlib's default value.
            constraint_name_fontrotation (float | None): the rotation angle of the constraint name in degree.The default uses matplotlib's default value.
            ylabel (str | None): the ylabel of figure. Defaults to "constraint violations".
            ylabel_size (float | None): the fontsize of the ylabel. The default uses matplotlib's default value.
            yticks (list[int | float] | None): the yticks of figure. Default to only integers by`MaxNLocator(integer=True)`.
            **matplotlib_boxplot_kwargs (dict): the parameter passed to matplotlib.boxplot. Defaults to `{showmeans: True, whis: [0, 100]}`.

        Returns:
            tuple[tuple[matplotlib.figure.Figure, matplotlib.axes.Subplot]]: A tuple of length equal to the number of rows in result. each element of is a tuple of figure and axes.


        Example:
            Below is the code to boxplot the constraint violations.
            In the first example, postprocessing the figure and axes changes the appearance of the plot.

            ```python
            import jijbench as jb
            import jijzept as jz
            from jijzept.sampler.openjij.sa_cpu import JijSAParameters
            from jijbench.visualization import ConstraintPlot

            problem = jb.get_problem("TSP")
            instance_data = jb.get_instance_data("TSP")[0][1]
            multipliers1 = {"onehot_time": 0.003, "onehot_location": 0.003}
            multipliers2 = {"onehot_time": 0.3, "onehot_location": 0.3}

            config_path = "XX"
            sa_parameter = JijSAParameters(num_reads=15)
            sa_sampler = jz.JijSASampler(config=config_path)

            bench = jb.Benchmark(
                params = {
                    "model": [problem],
                    "feed_dict": [instance_data],
                    "parameters": [sa_parameter],
                    "multipliers": [multipliers1, multipliers2],
                },
                solver = [sa_sampler.sample_model],
            )
            result = bench()

            cplot = ConstraintPlot(result)
            fig_ax_tuple = cplot.boxplot_violations()

            # you can post-process figure and axes to change the appearance of the plot.
            for fig, ax in fig_ax_tuple:
                fig.suptitle("my title")
                display(fig)
            ```

            By using the `construct_experiment_from_samplesets function`,
            `boxplot_violations` can also be used for `jm.SampleSet` obtained without `JijBenchmark`.

            ```python
            import jijbench as jb
            import jijzept as jz
            from jijbench.visualization import ConstraintPlot
            from jijbench.visualization import construct_experiment_from_samplesets

            problem = jb.get_problem("TSP")
            instance_data = jb.get_instance_data("TSP")[0][1]
            multipliers = {"onehot_time": 0.003, "onehot_location": 0.003}

            config_path = "XX"
            sampler = jz.JijSASampler(config=config_path)
            sampleset = sampler.sample_model(model=problem, feed_dict=instance_data, multipliers=multipliers, search=False, num_reads=100)

            result = construct_experiment_from_samplesets(sampleset)
            cplot = ConstraintPlot(result)
            fig_ax_tuple = cplot.boxplot_violations()
            ```
        """
        if ylabel is None:
            ylabel = "constraint violations"
        if len(matplotlib_boxplot_kwargs) == 0:
            # Show the arithmetic means in boxplot.
            matplotlib_boxplot_kwargs["showmeans"] = True
            # Make boxplot whisker positions min and max.
            matplotlib_boxplot_kwargs["whis"] = [0, 100]
        fig_ax_tuple = self.baseplot.boxplot(
            f=_get_violations_dict,
            figsize=figsize,
            title=title,
            title_fontsize=title_fontsize,
            xticklabels_size=constraint_name_fontsize,
            xticklabels_rotation=constraint_name_fontrotation,
            ylabel=ylabel,
            ylabel_size=ylabel_size,
            yticks=yticks,
            **matplotlib_boxplot_kwargs,
        )

        # Add a horizontal line to indicate that the constraint is satisfied. (violation = 0)
        fig_ax_list = []
        for fig, ax in fig_ax_tuple:
            ax.axhline(0, xmin=0, xmax=1, color="gray", linestyle="dotted")
            fig_ax_list.append((fig, ax))
        return tuple(fig_ax_list)



from __future__ import annotations

import jijmodeling as jm


def bin_packing():
    w = jm.Placeholder("w", dim=1)
    num_items = jm.Placeholder("n")
    c = jm.Placeholder("c")

    # y[j]: bin j を使用するかしないか
    y = jm.Binary("y", shape=(num_items,))
    # x[i][j]: item i を bin j に入れるとき1
    x = jm.Binary("x", shape=(num_items, num_items))

    # i: itemの添字
    i = jm.Element("i", num_items)
    # j: binの添字
    j = jm.Element("j", num_items)

    problem = jm.Problem("bin-packing")

    # objective function
    obj = y[:]
    problem += obj

    # Constraint1: 各itemをちょうど1つのbinにぶち込む
    const1 = jm.Constraint("onehot-constraint", jm.Sum(j, x[i, j]) - 1 == 0, forall=i)
    problem += const1

    # Constraint2: knapsack制約
    const2 = jm.Constraint(
        "knapsack-constraint", jm.Sum(i, w[i] * x[i, j]) - y[j] * c <= 0, forall=j
    )
    problem += const2

    return problem



from __future__ import annotations

import datetime
import typing as tp

import jijmodeling as jm
import numpy as np
import pandas as pd
from numpy.typing import NDArray

if tp.TYPE_CHECKING:
    from typing_extensions import TypeAlias

    from jijbench.containers.containers import Artifact, Record, Table
    from jijbench.experiment.experiment import Experiment
    from jijbench.node.base import DataNode
    from jijbench.solver.jijzept import SampleSet


# node
T = tp.TypeVar("T")
DataNodeInT = tp.TypeVar("DataNodeInT", bound="DataNode[tp.Any]")
DataNodeOutT = tp.TypeVar("DataNodeOutT", bound="DataNode[tp.Any]")
ConcatableT = tp.TypeVar(
    "ConcatableT", "Artifact", "Experiment", "Record", "SampleSet", "Table"
)


# element
ArrayType: TypeAlias = NDArray[tp.Union[np.int64, np.float64]]
DateTypes: TypeAlias = tp.Union[str, datetime.datetime, pd.Timestamp]
NumberTypes: TypeAlias = tp.Union[int, float, np.int64, np.float64]

# solver
ModelType: TypeAlias = tp.Tuple[jm.Problem, jm.PH_VALUES_INTERFACE]


# containers
ArtifactKeyType: TypeAlias = tp.Hashable
ArtifactValueType: TypeAlias = tp.Dict[tp.Hashable, "DataNode[tp.Any]"]
ArtifactDataType: TypeAlias = tp.Dict[ArtifactKeyType, ArtifactValueType]
ExperimentDataType: TypeAlias = tp.Tuple["Artifact", "Table"]



from __future__ import annotations

import glob
import pathlib
import typing as tp

import jijmodeling as jm
import rapidjson

from jijbench.datasets.problems import (
    bin_packing,
    knapsack,
    nurse_scheduling,
    travelling_salesman,
    travelling_salesman_with_time_windows,
)


class InstanceDataFileStorage:
    base_dir = pathlib.Path(__file__).parent.joinpath("Instances")

    def __init__(
        self,
        problem_name: tp.Literal[
            "bin-packing",
            "knapsack",
            "nurse-scheduling",
            "travelling-salesman",
            "travelling-salesman-with-time-windows",
        ],
    ):
        self.problem_name = problem_name

    def get_names(
        self,
        size: tp.Literal["small", "medium", "large"] = "small",
        n: int | tp.Literal["all"] = 1,
    ) -> list[str]:
        return list(map(lambda x: pathlib.Path(x).name, self.get_files(size, n)))

    def get_files(
        self,
        size: tp.Literal["small", "medium", "large"] = "small",
        n: int | tp.Literal["all"] = 1,
    ) -> list[str]:
        files = glob.glob(
            f"{self.base_dir}/{size}/{self.problem_name}/**/*.json", recursive=True
        )
        files.sort()
        if n == "all":
            return files
        elif isinstance(n, int):
            return files[:n]
        else:
            raise TypeError(f"n must be 'all' or int, but {type(n)} is given.")

    def get_files_map(
        self,
        size: tp.Literal["small", "medium", "large"] = "small",
        n: int | tp.Literal["all"] = 1,
    ) -> dict[str, str]:
        return {pathlib.Path(file).name: file for file in self.get_files(size, n)}

    def get_instance_data(
        self,
        size: tp.Literal["small", "medium", "large"] = "small",
        n: int | tp.Literal["all"] = 1,
    ) -> list[jm.PH_VALUES_INTERFACE]:
        return [self.load(file) for file in self.get_files(size, n)]

    @staticmethod
    def load(file: str) -> jm.PH_VALUES_INTERFACE:
        if not pathlib.Path(file).exists():
            raise FileNotFoundError(f"'{file}' is not found.")

        with open(file, "r") as f:
            instance_data = rapidjson.load(f)
        return instance_data


def get_models(
    problem_name: tp.Literal[
        "bin-packing",
        "knapsack",
        "nurse-scheduling",
        "travelling-salesman",
        "travelling-salesman-with-time-windows",
    ],
    size: tp.Literal["small", "medium", "large"] = "small",
    n: int | tp.Literal["all"] = 1,
) -> list[tuple[jm.Problem, jm.InstanceData]]:
    storage = InstanceDataFileStorage(problem_name)
    problem = get_problem(problem_name)
    files = storage.get_instance_data(size, n)
    return list(zip([problem] * n, files))


def get_problem(
    problem_name: tp.Literal[
        "bin-packing",
        "knapsack",
        "nurse-scheduling",
        "travelling-salesman",
        "travelling-salesman-with-time-windows",
    ]
) -> jm.Problem:
    if problem_name == "bin-packing":
        return bin_packing()
    elif problem_name == "knapsack":
        return knapsack()
    elif problem_name == "nurse-scheduling":
        return nurse_scheduling()
    elif problem_name == "travelling-salesman":
        return travelling_salesman()
    elif problem_name == "travelling-salesman-with-time-windows":
        return travelling_salesman_with_time_windows()
    else:
        raise ValueError(f"Unknown problem name: {problem_name}")


def get_instance_data(
    problem_name: tp.Literal[
        "bin-packing",
        "knapsack",
        "nurse-scheduling",
        "travelling-salesman",
        "travelling-salesman-with-time-windows",
    ],
    size: tp.Literal["small", "medium", "large"] = "small",
    n: int | tp.Literal["all"] = 1,
) -> list[jm.PH_VALUES_INTERFACE]:
    storage = InstanceDataFileStorage(problem_name)
    return storage.get_instance_data(size, n)



from pkgutil import extend_path

__path__ = extend_path(__path__, __name__)

import jijbench.functions as functions
import jijbench.node as node
from jijbench.benchmark.benchmark import Benchmark, benchmark_for
from jijbench.benchmark.decorator import checkpoint
from jijbench.containers.containers import Artifact, Record, Table
from jijbench.datasets.model import get_instance_data, get_models, get_problem
from jijbench.elements.array import Array
from jijbench.elements.base import Callable, Number, String
from jijbench.elements.date import Date
from jijbench.elements.id import ID
from jijbench.evaluation.evaluation import Evaluation
from jijbench.experiment.experiment import (
    Experiment,
    get_benchmark_ids,
    get_experiment_ids,
    get_id_table,
)
from jijbench.io.io import load, save
from jijbench.solver.base import Parameter, Response, Solver
from jijbench.solver.jijzept import InstanceData, Model, SampleSet

__all__ = [
    "checkpoint",
    "benchmark_for",
    "functions",
    "node",
    "get_benchmark_ids",
    "get_experiment_ids",
    "get_id_table",
    "get_instance_data",
    "get_models",
    "get_problem",
    "load",
    "save",
    "Array",
    "Artifact",
    "Benchmark",
    "Callable",
    "Date",
    "Evaluation",
    "Experiment",
    "ID",
    "InstanceData",
    "Model",
    "Number",
    "Parameter",
    "Record",
    "Response",
    "SampleSet",
    "Solver",
    "Table",
    "String",
]



from __future__ import annotations

import abc
import datetime
import pathlib
import typing as tp
import uuid
import warnings
from dataclasses import dataclass, field

import numpy as np
import pandas as pd

from jijbench.consts.default import DEFAULT_RESULT_DIR
from jijbench.containers.containers import Artifact, Container, Record, Table
from jijbench.elements.id import ID
from jijbench.functions.concat import Concat
from jijbench.functions.factory import ArtifactFactory, TableFactory
from jijbench.io.io import save
from jijbench.solver.base import Parameter, Response
from jijbench.typing import ArtifactDataType, ExperimentDataType


@dataclass
class Experiment(Container[ExperimentDataType]):
    """Stores data related to an benchmark.

    The Experiment class stores the results obtained from a benchmark as Artifact and Table objects and assists in managing the benchmark process.
    With this class, you can add and save experimental results, as well as view them in various formats.

    Attributes:
        data (tuple[Artifact, Table]): A tuple containing an Artifact object and a Table object.
        name (str): The name of the experiment.
        autosave (bool): Whether to automatically save the experiment upon exit.
        savedir (str | pathlib.Path): The directory where the experiment will be saved.
    """

    data: tuple[Artifact, Table] = field(default_factory=lambda: (Artifact(), Table()))
    name: str = field(default_factory=lambda: str(uuid.uuid4()))
    autosave: bool = field(default=True, repr=False)
    savedir: str | pathlib.Path = field(default=DEFAULT_RESULT_DIR, repr=False)

    def __post_init__(self):
        super().__post_init__()
        self.savedir = pathlib.Path(self.savedir)
        setattr(self, "state", _Created())

    def __len__(self) -> int:
        """
        Perform the operation __len__.
        """
        a_len = len(self.data[0])
        t_len = len(self.data[1])
        if a_len != t_len:
            warnings.warn(
                f"The length of artifact object and table object are different: {a_len} != {t_len}, return the larger one."
            )
        return max(a_len, t_len)

    def __enter__(self) -> Experiment:
        """Set up Experiment.
        Automatically makes a directory for saving the experiment, if it doesn't exist.
        """

        setattr(self, "state", _Running(self.name))
        pathlib.Path(self.savedir).mkdir(parents=True, exist_ok=True)
        return self

    def __exit__(self, exception_type, exception_value, traceback) -> None:
        """Saves the experiment if autosave is True."""
        state = getattr(self, "state")

        if self.autosave:
            state.save(self)

        setattr(self, "state", _Waiting(self.name))

    @property
    def artifact(self) -> ArtifactDataType:
        """Return the artifact of the experiment as a dictionary."""
        return self.data[0].view()

    @property
    def table(self) -> pd.DataFrame:
        """Return the table of the experiment as a pandas dataframe."""
        return self.data[1].view()

    @property
    def params_table(self) -> pd.DataFrame:
        """Return the parameters table of the experiment as a pandas dataframe."""
        _, t = self.data
        bools = t.data.apply(lambda x: not isinstance(x[0], Parameter))
        droped_columns = t.data.columns[bools]
        data = t.data.drop(columns=droped_columns)
        return Table(data, self.name).view()

    @property
    def response_table(self) -> pd.DataFrame:
        """Return the returns table of the experiment as a pandas dataframe."""
        _, t = self.data
        bools = t.data.apply(lambda x: not isinstance(x[0], Response))
        droped_columns = t.data.columns[bools]
        data = t.data.drop(columns=droped_columns)
        return Table(data, self.name).view()

    @classmethod
    def validate_data(cls, data: ExperimentDataType) -> ExperimentDataType:
        """Validate the data of the experiment.

        Args:
            data (ExperimentDataType): The data to validate.

        Raises:
            TypeError: If data is not an instance of ExperimentDataType or
            if the first element of data is not an instance of Artifact or
            if the second element of data is not an instance of Table.

        Returns:
            ExperimentDataType: The validated data.
        """
        if not isinstance(data, tuple):
            raise TypeError(f"Data must be a tuple, got {type(data)}.")

        if len(data) != 2:
            raise ValueError(f"Data must be a tuple of length 2, got {len(data)}.")

        artifact, table = data
        if not isinstance(artifact, Artifact):
            raise TypeError(
                f"Type of attribute data is {ExperimentDataType}, and data[0] must be Artifact instead of {type(artifact).__name__}."
            )
        if not isinstance(table, Table):
            raise TypeError(
                f"Type of attribute data is {ExperimentDataType}, and data[1] must be Table instead of {type(artifact).__name__}."
            )
        return data

    def view(self) -> tuple[dict, pd.DataFrame]:
        """Return a tuple of the artifact dictionary and table dataframe."""
        return (self.data[0].view(), self.data[1].view())

    def append(self, record: Record) -> None:
        """Append a new record to the experiment.

        Args:
            record (Record): The record to be appended to the experiment.
        """
        state = getattr(self, "state")
        state.append(self, record)

    def star(self) -> None:
        """Mark the experiment."""

        savedir = pathlib.Path(self.savedir)
        benchmark_id = None
        if (savedir / "star.csv").exists():
            star_file = savedir / "star.csv"
        elif (savedir.parent / "star.csv").exists():
            star_file = savedir.parent / "star.csv"
            benchmark_id = savedir.name
        else:
            star_file = savedir / "star.csv"

        if star_file.exists():
            if star_file.stat().st_size:
                star = pd.read_csv(star_file, index_col=0)
            else:
                star = pd.DataFrame(
                    columns=["benchmark_id", "experiment_id", "savedir"]
                )
            star.loc[len(star), ["benchmark_id", "experiment_id", "savedir"]] = [
                benchmark_id or np.nan,
                self.name,
                self.savedir,
            ]
            star = star.drop_duplicates(subset=["experiment_id"])
        else:
            star_file.parent.mkdir(parents=True, exist_ok=True)
            star = pd.DataFrame(
                {
                    "benchmark_id": benchmark_id or np.nan,
                    "experiment_id": [self.name],
                    "savedir": [self.savedir],
                }
            )
        star.to_csv(star_file)

    def save(self):
        """Save the experiment."""
        pathlib.Path(self.savedir).mkdir(parents=True, exist_ok=True)
        state = getattr(self, "state")
        state.save(self)


class _ExperimentState(metaclass=abc.ABCMeta):
    def __init__(self, index: tp.Hashable = 0) -> None:
        self.index = index

    @abc.abstractmethod
    def append(self, context: Experiment, record: Record) -> None:
        pass

    @abc.abstractmethod
    def save(self, context: Experiment) -> None:
        pass


class _Created(_ExperimentState):
    def append(self, context: Experiment, record: Record) -> None:
        record.name = self.index
        _append(context, record)
        if isinstance(self.index, int):
            self.index += 1
        context.state = _Waiting(self.index)

    def save(self, context: Experiment) -> None:
        save(context, savedir=context.savedir)


class _Waiting(_ExperimentState):
    def append(self, context: Experiment, record: Record) -> None:
        record.name = self.index
        _append(context, record)
        if isinstance(self.index, int):
            self.index += 1

    def save(self, context: Experiment) -> None:
        save(context, savedir=context.savedir)


class _Running(_ExperimentState):
    def append(self, context: Experiment, record: Record) -> None:
        run_id = ID().data
        if isinstance(self.index, str):
            record.name = (self.index, run_id)
        elif isinstance(self.index, tp.Iterable):
            record.name = (*self.index, run_id)
        else:
            record.name = (self.index, run_id)
        _append(context, record)

    def save(
        self,
        context: Experiment,
    ) -> None:
        a, t = context.data
        latest = t.index[-1]

        ai = Artifact({latest: a.data[latest]}, a.name)
        ti = Table(t.data.loc[[latest]], t.name)

        experiment = Experiment(
            (ai, ti), context.name, autosave=context.autosave, savedir=context.savedir
        )
        save(experiment, savedir=context.savedir, mode="a")


def _append(context: Experiment, record: Record) -> None:
    concat: Concat[Experiment] = Concat()
    data = (ArtifactFactory()([record]), TableFactory()([record]))
    other = type(context)(
        data, context.name, autosave=context.autosave, savedir=context.savedir
    )
    node = context.apply(
        concat,
        [other],
        name=context.name,
        autosave=context.autosave,
        savedir=context.savedir,
    )
    context._update_attrs(node)


def get_benchmark_ids(
    savedir: str | pathlib.Path = DEFAULT_RESULT_DIR,
) -> list[str | float]:
    """Get the benchmark ids.

    Args:
        savedir (str | pathlib.Path, optional): The directory to save the experiment. Defaults to DEFAULT_RESULT_DIR.

    Returns:
        list[str]: The benchmark ids.
    """

    benchmark_ids: list[str | float] = []
    for d in pathlib.Path(savedir).glob("*"):
        if d.is_dir():
            if list(d.glob("artifact")):
                benchmark_ids.append(np.nan)
            else:
                benchmark_ids.extend([d.name] * len(get_experiment_ids(d)))

    return benchmark_ids


def get_experiment_ids(
    savedir: str | pathlib.Path = DEFAULT_RESULT_DIR, only_star: bool = False
) -> list[str]:
    """Get the experiment ids.

    Args:
        savedir (str | pathlib.Path, optional): The directory to save the experiment. Defaults to DEFAULT_RESULT_DIR.
        only_star (bool, optional): Whether to only return the starred experiments. Defaults to False.

    Returns:
        list[str]: The experiment ids.
    """
    if only_star:
        star_file = pathlib.Path(savedir) / "star.csv"
        if star_file.exists():
            return pd.read_csv(star_file, index_col=0)["experiment_id"].tolist()
        else:
            return []
    else:
        experiment_ids: list[str] = []
        for d in pathlib.Path(savedir).glob("*"):
            if d.is_dir():
                if list(d.glob("artifact")):
                    experiment_ids.append(d.name)
                else:
                    experiment_ids.extend(get_experiment_ids(d))
        return experiment_ids


def get_id_table(
    savedir: str | pathlib.Path = DEFAULT_RESULT_DIR,
) -> pd.DataFrame:
    """Get the experiment info.

    Args:
        savedir (str | pathlib.Path, optional): The directory to save the experiment. Defaults to DEFAULT_RESULT_DIR.

    Returns:
        pd.DataFrame: The experiment info.
    """

    info = pd.DataFrame(
        {
            "benchmark_id": get_benchmark_ids(savedir),
            "experiment_id": get_experiment_ids(savedir),
        }
    )

    info["savedir"] = info.apply(
        lambda x: f"{savedir}/{x['benchmark_id']}/{x['experiment_id']}"
        if isinstance(x["benchmark_id"], str)
        else f"{savedir}/{x['experiment_id']}",
        axis=1,
    )

    info["timestamp"] = info["savedir"].apply(
        lambda x: datetime.datetime.fromtimestamp(
            pathlib.Path(x).stat().st_ctime
        ).strftime("%Y-%m-%d %H:%M:%S")
    )
    info["star"] = ""

    star_file = pathlib.Path(savedir) / "star.csv"
    if star_file.exists():
        star = pd.read_csv(star_file, index_col=0)
        mask = info["experiment_id"].isin(star["experiment_id"].tolist())
        info.loc[mask, "star"] = "*"
    return info



"""
This script provides a dashboard application using Streamlit to visualize and analyze experiment for optimization problem.
The dashboard consists of four tabs: "Instance data", "Problem", "Solver", and "Analysis",
allowing users to interact with and explore the benchmark data in various ways.

Usage:
To run the dashboard application, execute this script using a Python interpreter. Make sure that Streamlit
and other required packages are installed in your Python environment.

Attributes:
logdir (str): An optional environment variable specifying the directory containing the benchmark data.
If not provided, the default directory defined in DEFAULT_RESULT_DIR is used.

Functions:
run(): Initializes the dashboard application and sets up the tabs for user interaction.
"""

from __future__ import annotations

import os
import pathlib

import streamlit as st

from jijbench.consts.default import DEFAULT_RESULT_DIR
from jijbench.dashboard.session import Session

st.set_page_config(layout="wide")


def run():
    st.title("JB Board")

    logdir = pathlib.Path(os.environ.get("logdir", DEFAULT_RESULT_DIR))
    session = Session(logdir)

    tab_names = ["Analysis", "Instance data", "Problem", "Solver"]
    tab_map = {name: tab for name, tab in zip(tab_names, st.tabs(tab_names))}

    with tab_map["Analysis"]:
        session.display_page("Analysis")

    with tab_map["Instance data"]:
        session.display_page("Instance data")

    with tab_map["Problem"]:
        session.display_page("Problem")

    with tab_map["Solver"]:
        session.display_page("Solver")


if __name__ == "__main__":
    run()



from __future__ import annotations

import jijmodeling as jm


def knapsack():
    w = jm.Placeholder("w", dim=1)
    v = jm.Placeholder("v", dim=1)
    n = jm.Placeholder("n")
    c = jm.Placeholder("c")
    x = jm.Binary("x", shape=(n,))

    # i: itemの添字
    i = jm.Element("i", n)

    problem = jm.Problem("knapsack")

    # objective function
    obj = jm.Sum(i, v[i] * x[i])
    problem += -1 * obj

    # Constraint: knapsack 制約
    const = jm.Constraint("knapsack-constraint", jm.Sum(i, w[i] * x[i]) - c <= 0)
    problem += const

    return problem



from __future__ import annotations

import pathlib
import typing as tp

import jijmodeling as jm
import numpy as np
import pandas as pd

from jijbench.consts.default import DEFAULT_RESULT_DIR
from jijbench.elements.base import Any
from jijbench.elements.id import ID
from jijbench.node.base import DataNode, FunctionNode
from jijbench.typing import ConcatableT

if tp.TYPE_CHECKING:
    from typing_extensions import TypeGuard

    from jijbench.containers.containers import Artifact, Record, Table
    from jijbench.experiment.experiment import Experiment
    from jijbench.solver.jijzept import SampleSet


def _is_artifact_list(inputs: list[ConcatableT]) -> TypeGuard[list[Artifact]]:
    from jijbench.containers.containers import Artifact

    return all([isinstance(node, Artifact) for node in inputs])


def _is_experiment_list(inputs: list[ConcatableT]) -> TypeGuard[list[Experiment]]:
    from jijbench.experiment.experiment import Experiment

    return all([isinstance(node, Experiment) for node in inputs])


def _is_record_list(inputs: list[ConcatableT]) -> TypeGuard[list[Record]]:
    from jijbench.containers.containers import Record

    return all([isinstance(node, Record) for node in inputs])


def _is_table_list(inputs: list[ConcatableT]) -> TypeGuard[list[Table]]:
    from jijbench.containers.containers import Table

    return all([isinstance(node, Table) for node in inputs])


def _is_sampleset_list(inputs: list[ConcatableT]) -> TypeGuard[list[SampleSet]]:
    from jijbench.solver.jijzept import SampleSet

    return all([isinstance(node, SampleSet) for node in inputs])


def _is_datanode_list(inputs: list[ConcatableT]) -> bool:
    sample = inputs[0]
    is_datanode = isinstance(sample, DataNode)
    return all([isinstance(node, sample.__class__) for node in inputs]) & is_datanode


class Concat(FunctionNode[ConcatableT, ConcatableT]):
    """Concat class for concatenating multiple data nodes.

    This class can be apply to `Artifact`, `Experiment`, `Record`, `Table`, `SampleSet`.
    """

    @tp.overload
    def __call__(self, inputs: list[Artifact], name: tp.Hashable = None) -> Artifact:
        """Concatenate the gitven multiple 'Artifact' objects.

        Args:
            inputs (list[Artifact]): A list of 'Artifact', 'Experiment', 'Record' or 'Table' objects to concatenate.
            name (tp.Hashable, optional): The name of the resulting object. Defaults to None. Defaults to None.

        Raises:
            TypeError: If the type of elements in 'inputs' are not unified or if the 'name' attribute is not a string.

        Returns:
            Artifact: The resulting 'Artifact' object.
        """
        ...

    @tp.overload
    def __call__(
        self,
        inputs: list[Experiment],
        name: str | None = None,
        *,
        autosave: bool = True,
        savedir: str | pathlib.Path = DEFAULT_RESULT_DIR,
        axis: tp.Literal[0, 1] = 0,
        index_name: str | None = None,
    ) -> Experiment:
        """Concatenate the gitven multiple 'Experiment' objects.

        Args:
            inputs (list[Experiment]): A list of 'Artifact', 'Experiment', 'Record' or 'Table' objects to concatenate.
            name (tp.Hashable, optional): The name of the resulting object. Defaults to None. Defaults to None.
            autosave (bool, optional): If True, the resulting object will be saved to disk. Defaults to True.
            savedir (str | pathlib.Path, optional): The directory where the resulting object will be saved. Defaults to DEFAULT_RESULT_DIR.

        Raises:
            TypeError: If the type of elements in 'inputs' are not unified or if the 'name' attribute is not a string.

        Returns:
            Experiment: The resulting 'Experiment' object.
        """
        ...

    @tp.overload
    def __call__(self, inputs: list[Record], name: tp.Hashable = None) -> Record:
        """Concatenate the gitven multiple 'Record' objects.

        Args:
            inputs (list[Record]): A list of 'Artifact', 'Experiment', 'Record' or 'Table' objects to concatenate.
            name (tp.Hashable, optional): The name of the resulting object. Defaults to None. Defaults to None.

        Raises:
            TypeError: If the type of elements in 'inputs' are not unified or if the 'name' attribute is not a string.

        Returns:
            Record: The resulting 'Record' object.
        """
        ...

    @tp.overload
    def __call__(
        self,
        inputs: list[Table],
        name: tp.Hashable = None,
        *,
        axis: tp.Literal[0, 1] = 0,
        index_name: str | None = None,
    ) -> Table:
        """Concatenate the gitven multiple 'Artifact' objects.

        Args:
            inputs (list[Table]): A list of 'Artifact', 'Experiment', 'Record' or 'Table' objects to concatenate.
            name (tp.Hashable, optional): The name of the resulting object. Defaults to None. Defaults to None.
            axis (tp.Literal[0, 1], optional): The axis along which to concatenate the input 'Table' objects. Defaults to 0.
            index_name (str | None, optional): The name of the resulting object's index. Defaults to None.

        Raises:
            TypeError: If the type of elements in 'inputs' are not unified or if the 'name' attribute is not a string.

        Returns:
            Table: The resulting 'Table' object.
        """
        ...

    @tp.overload
    def __call__(self, inputs: list[SampleSet], name: str) -> SampleSet:
        """Concatenate the gitven multiple 'SampleSet' objects.

        Args:
            inputs (list[SampleSet]): A list of 'Artifact', 'Experiment', 'Record' or 'Table' objects to concatenate.
            name (tp.Hashable, optional): The name of the resulting object. Defaults to None. Defaults to None.

        Raises:
            TypeError: If the type of elements in 'inputs' are not unified or if the 'name' attribute is not a string.

        Returns:
            SampleSet: The resulting 'SampleSet' object.
        """
        ...

    def __call__(
        self,
        inputs: list[tp.Any],
        name: tp.Hashable = None,
        *,
        autosave: bool = True,
        savedir: str | pathlib.Path = DEFAULT_RESULT_DIR,
        axis: tp.Literal[0, 1] = 0,
        index_name: str | None = None,
    ) -> tp.Any:
        """Concatenate Given a list of either 'Artifact', 'Experiment', 'Record', or 'Table' objects."""
        if _is_datanode_list(inputs):
            return super().__call__(
                inputs,
                name=name,
                autosave=autosave,
                savedir=savedir,
                axis=axis,
                index_name=index_name,
            )
        else:
            raise TypeError(
                "Type of elements in 'inputs' must be unified either 'Artifact', 'Experiment', 'Record' or 'Table'."
            )

    @tp.overload
    def operate(self, inputs: list[Artifact], name: tp.Hashable = None) -> Artifact:
        """Concatenate the gitven multiple 'Artifact' objects.

        Args:
            inputs (list[Artifact]): A list of 'Artifact', 'Experiment', 'Record' or 'Table' objects to concatenate.
            name (tp.Hashable, optional): The name of the resulting object. Defaults to None. Defaults to None.

        Raises:
            TypeError: If the type of elements in 'inputs' are not unified or if the 'name' attribute is not a string.

        Returns:
            Artifact: The resulting 'Artifact' object.
        """
        ...

    @tp.overload
    def operate(
        self,
        inputs: list[Experiment],
        name: str | None = None,
        *,
        autosave: bool = True,
        savedir: str | pathlib.Path = DEFAULT_RESULT_DIR,
    ) -> Experiment:
        """Concatenate the gitven multiple 'Experiment' objects.

        Args:
            inputs (list[Experiment]): A list of 'Artifact', 'Experiment', 'Record' or 'Table' objects to concatenate.
            name (tp.Hashable, optional): The name of the resulting object. Defaults to None. Defaults to None.
            autosave (bool, optional): If True, the resulting object will be saved to disk. Defaults to True.
            savedir (str | pathlib.Path, optional): The directory where the resulting object will be saved. Defaults to DEFAULT_RESULT_DIR.

        Raises:
            TypeError: If the type of elements in 'inputs' are not unified or if the 'name' attribute is not a string.

        Returns:
            Experiment: The resulting 'Experiment' object.
        """
        ...

    @tp.overload
    def operate(self, inputs: list[Record], name: tp.Hashable = None) -> Record:
        """Concatenate the gitven multiple 'Record' objects.

        Args:
            inputs (list[Record]): A list of 'Artifact', 'Experiment', 'Record' or 'Table' objects to concatenate.
            name (tp.Hashable, optional): The name of the resulting object. Defaults to None. Defaults to None.

        Raises:
            TypeError: If the type of elements in 'inputs' are not unified or if the 'name' attribute is not a string.

        Returns:
            Record: The resulting 'Record' object.
        """
        ...

    @tp.overload
    def operate(
        self,
        inputs: list[Table],
        name: tp.Hashable = None,
        *,
        axis: tp.Literal[0, 1] = 0,
        index_name: str | None = None,
    ) -> Table:
        """Concatenate the gitven multiple 'Artifact' objects.

        Args:
            inputs (list[Table]): A list of 'Artifact', 'Experiment', 'Record' or 'Table' objects to concatenate.
            name (tp.Hashable, optional): The name of the resulting object. Defaults to None. Defaults to None.
            axis (tp.Literal[0, 1], optional): The axis along which to concatenate the input 'Table' objects. Defaults to 0.
            index_name (str | None, optional): The name of the resulting object's index. Defaults to None.

        Raises:
            TypeError: If the type of elements in 'inputs' are not unified or if the 'name' attribute is not a string.

        Returns:
            Table: The resulting 'Table' object.
        """
        ...

    @tp.overload
    def operate(self, inputs: list[SampleSet], name: str) -> SampleSet:
        """Concatenate the gitven multiple 'SampleSet' objects.

        Args:
            inputs (list[SampleSet]): A list of 'Artifact', 'Experiment', 'Record' or 'Table' objects to concatenate.
            name (tp.Hashable, optional): The name of the resulting object. Defaults to None. Defaults to None.

        Raises:
            TypeError: If the type of elements in 'inputs' are not unified or if the 'name' attribute is not a string.

        Returns:
            SampleSet: The resulting 'SampleSet' object.
        """
        ...

    def operate(
        self,
        inputs: list[tp.Any],
        name: tp.Hashable = None,
        *,
        autosave: bool = True,
        savedir: str | pathlib.Path = DEFAULT_RESULT_DIR,
        axis: tp.Literal[0, 1] = 0,
        index_name: str | None = None,
    ) -> tp.Any:
        """Concatenate Given a list of either 'Artifact', 'Experiment', 'Record', or 'Table' objects."""
        if _is_artifact_list(inputs):
            data = {}
            for node in inputs:
                for k, v in node.data.items():
                    if k in data:
                        data[k].update(v.copy())
                    else:
                        data[k] = v.copy()
            return type(inputs[0])(data, name)
        elif _is_experiment_list(inputs):
            concat_a: Concat[Artifact] = Concat()
            concat_t: Concat[Table] = Concat()
            inputs_a = [n.data[0] for n in inputs]
            inputs_t = [n.data[1] for n in inputs]
            artifact = inputs_a[0].apply(concat_a, inputs_a[1:])
            table = inputs_t[0].apply(
                concat_t, inputs_t[1:], axis=axis, index_name=index_name
            )

            if name is None:
                name = ID().data

            if not isinstance(name, str):
                raise TypeError("Attirbute name of Experiment must be string.")

            return type(inputs[0])(
                (artifact, table),
                name,
                autosave=autosave,
                savedir=savedir,
            )
        elif _is_record_list(inputs):
            data = pd.concat([node.data for node in inputs])
            return type(inputs[0])(data, name)
        elif _is_table_list(inputs):
            data = pd.concat([node.data for node in inputs], axis=axis)
            data = data.fillna(Any(np.nan, "NaN"))
            data.index.name = index_name
            return type(inputs[0])(data, name)
        elif _is_sampleset_list(inputs):
            data = jm.concatenate([node.data for node in inputs])
            return type(inputs[0])(data, str(name))
        else:
            raise TypeError(
                "Type of elements in 'inputs' must be unified either 'Artifact', 'Experiment', 'Record' or 'Table'."
            )



from __future__ import annotations

import codecs
import datetime
import glob
import pathlib
import re
import sys
import typing as tp

import jijmodeling as jm
import matplotlib
import numpy as np
import pandas as pd
import plotly.express as px
import plotly.graph_objs as go
import rapidjson
import streamlit as st
from plotly.colors import n_colors
from st_aggrid import AgGrid, ColumnsAutoSizeMode, GridUpdateMode
from st_aggrid.grid_options_builder import GridOptionsBuilder
from streamlit.delta_generator import DeltaGenerator
from streamlit_ace import st_ace
from streamlit_elements import editor, elements
from streamlit_tree_select import tree_select
from typeguard import check_type
from typing_extensions import TypeGuard

import jijbench as jb

if tp.TYPE_CHECKING:
    from jijbench.dashboard.session import Session


class RoutingHandler:
    """
    This class provides methods to handle the navigation between different pages of the application,
    such as instance data selection, solver configuration, problem definition, and result analysis.
    """

    def on_select_page(self, session: Session) -> None:
        """
        Handle the navigation to the selected page.

        Args:
            session (Session): The current session.
        """
        page = session.state.selected_page
        if page == "Instance data":
            self.on_select_instance_data(session)
        elif page == "Problem":
            self.on_select_problem(session)
        elif page == "Solver":
            self.on_select_solver(session)
        elif page == "Analysis":
            self.on_select_analysis(session)

    def on_select_instance_data(self, session: Session) -> None:
        """
        Display the instance data selection and visualization options.

        Args:
            session (Session): The current session.
        """
        session.state.selected_figure_for_instance_data = st.radio(
            "Fugure",
            options=["Histogram", "Box", "Violin"],
            horizontal=True,
        )
        options = sum(
            [
                [
                    f"{problem_name}/{pathlib.Path(f).name}"
                    for f in session.state.selected_instance_data_files
                    if problem_name in f
                ]
                for problem_name in session.state.selected_problem_names
            ],
            [],
        )
        session.state.selected_instance_data_name = st.radio(
            "Loaded instance data",
            options=options,
            horizontal=True,
        )

        ph_plot = st.empty()

        cols = st.columns(2)
        with cols[0]:
            with st.expander("List", expanded=True):
                session.state.selected_instance_data_map = tree_select(
                    session.state.instance_data_dir_tree.nodes,
                    check_model="leaf",
                    only_leaf_checkboxes=True,
                )
                if st.button("Load", key="load_instance_data"):
                    session.state.is_instance_data_loaded = True
                    with ph_plot.container():
                        session.plot_instance_data()

        with cols[1]:
            with st.expander("Upload", expanded=True):
                with st.form("new_instance_data"):
                    session.state.input_problem_name = st.text_input(
                        "Input problem name"
                    )
                    byte_stream = st.file_uploader(
                        "Upload your instance data", type=["json"]
                    )
                    if byte_stream:
                        session.state.uploaded_instance_data_name = byte_stream.name
                        ins_d = rapidjson.loads(byte_stream.getvalue())
                    if st.form_submit_button("Submit"):
                        if byte_stream:
                            session.add_instance_data()
                            st.experimental_rerun()

    def on_select_solver(self, session: Session) -> None:
        """
        Display the solver selection and configuration options.

        Args:
            session (Session): The current session.
        """
        st.info("Coming soon...")

    def on_select_problem(self, session: Session) -> None:
        """
        Display the problem definition and visualization options.

        Args:
            session (Session): The current session.
        """

        def is_callable(obj: tp.Any, name: str) -> TypeGuard[tp.Callable[..., tp.Any]]:
            check_type(name, obj, tp.Callable[..., tp.Any])
            return True

        def get_function_from_code(code: str) -> tp.Callable[..., tp.Any]:
            module = sys.modules[__name__]
            func_name = code.split("(")[0].split("def ")[-1]
            exec(code, globals())
            func = getattr(module, func_name)
            if is_callable(func, func_name):
                return func
            else:
                raise TypeError("The code must be function format.")

        # Aceエディタの初期値
        initial_code = """def your_problem():\n\t..."""
        # Aceエディタの設定
        editor_options = {
            "value": initial_code,
            "placeholder": "",
            "height": "300px",
            "language": "python",
            "theme": "ambiance",
            "keybinding": "vscode",
            "min_lines": 12,
            "max_lines": None,
            "font_size": 12,
            "tab_size": 4,
            "wrap": False,
            "show_gutter": True,
            "show_print_margin": False,
            "readonly": False,
            "annotations": None,
            "markers": None,
            "auto_update": False,
            "key": None,
        }

        code = codecs.decode(st_ace(**editor_options), "unicode_escape")

        if st.button("Run"):
            st.info("Coming soon...")
            # func = get_function_from_code(code)
            # problem = func()
            # st.latex(problem._repr_latex_()[2:-2])

    def on_select_analysis(self, session: Session) -> None:
        """
        Display the benchmark results and analysis options.

        Args:
            session (Session): The current session.
        """

        components = {
            "id_table": st.empty(),
            "result_table": st.empty(),
            "evaluation_table": st.empty(),
            "problem": st.empty(),
            "metric_plot": st.empty(),
            "sampleset_analysis": st.empty(),
            "scatter": st.empty(),
            "parallel_coordinates": st.empty(),
        }
        plot_components: dict[str, DeltaGenerator] = {}

        with components["id_table"].container():
            st.subheader("Experiment history")
            id_table = jb.get_id_table(savedir=session.state.logdir)
            id_table = (
                id_table.sort_values("timestamp", ascending=False)
                .replace("*", "⭐️")
                .drop(columns=["savedir"])
            )
            id_table.insert(0, "", [""] * len(id_table))

            gob = GridOptionsBuilder.from_dataframe(id_table)
            gob.configure_selection(use_checkbox=True, selection_mode="multiple")
            gob.configure_column("star", cellStyle={"textAlign": "center"})
            grid_options = gob.build()
            grid_id_table = AgGrid(
                id_table,
                height=250,
                gridOptions=grid_options,
                grid_update_mode=GridUpdateMode.VALUE_CHANGED,
                columns_auto_size_mode=ColumnsAutoSizeMode.FIT_CONTENTS,
            )
            session.state.selected_benchmark_results = grid_id_table["selected_rows"]
            is_benchmark_results_loaded = st.button(
                "Load", key="load_benchmark_results"
            )
            if is_benchmark_results_loaded:
                session.state.num_experiment_loaded += 1

        if session.state.num_experiment_loaded:
            benchmark_ids = session.state.selected_benchmark_ids
            params_table = _get_params_table(benchmark_ids, session.state.logdir)

            with components["result_table"].container():
                st.subheader("Table")
                response_table = _get_response_table(
                    benchmark_ids, session.state.logdir
                )
                table = pd.concat([params_table, response_table], axis=1)
                st.dataframe(table)
                gob = GridOptionsBuilder.from_dataframe(table)
                gob.configure_columns(
                    params_table.columns,
                    cellStyle={"backgroundColor": "#f8f9fb"},
                    pinned="left",
                )
                grid_options = gob.build()
                AgGrid(
                    table,
                    height=500,
                    gridOptions=grid_options,
                    grid_update_mode=GridUpdateMode.VALUE_CHANGED,
                    columns_auto_size_mode=ColumnsAutoSizeMode.FIT_CONTENTS,
                )
                st.markdown("<hr>", unsafe_allow_html=True)

            with components["evaluation_table"].container():
                st.subheader("Evaluation")
                cols = st.columns(8)
                with cols[0]:
                    opt_value = st.text_input("opt value", value=None)
                    if _is_number_str(opt_value):
                        opt_value = float(opt_value)
                    else:
                        opt_value = None
                with cols[1]:
                    pr = st.text_input(
                        "pr",
                        value=0.99,
                    )
                    if _is_number_str(pr):
                        pr = float(pr)
                evaluation_table = _get_evaluation_table(
                    benchmark_ids, session.state.logdir
                )
                table = pd.concat([params_table, evaluation_table], axis=1)
                gob = GridOptionsBuilder.from_dataframe(table)
                gob.configure_columns(
                    params_table.columns,
                    cellStyle={"backgroundColor": "#f8f9fb"},
                    pinned="left",
                )
                grid_options = gob.build()
                AgGrid(
                    table,
                    height=500,
                    gridOptions=grid_options,
                    grid_update_mode=GridUpdateMode.VALUE_CHANGED,
                    columns_auto_size_mode=ColumnsAutoSizeMode.FIT_CONTENTS,
                )
                st.markdown("<hr>", unsafe_allow_html=True)

            with components["problem"].container():
                st.subheader("Problems")

                def get_unique_problems(x: pd.Series) -> pd.Series:
                    problem_names = x.apply(lambda x: x.name)
                    problem_names.name = f"{x.name}_name"
                    return pd.concat([x, problem_names], axis=1).drop_duplicates(
                        f"{x.name}_name"
                    )[x.name]

                problem_table = _get_problem_table(benchmark_ids, session.state.logdir)
                problems = np.unique(
                    [
                        p
                        for problems in problem_table.apply(get_unique_problems).values
                        for p in problems
                    ]
                ).tolist()
                for problem in problems:
                    with st.expander(f"{problem.name}"):
                        st.latex(problem._repr_latex_()[2:-2])
                st.markdown("<hr>", unsafe_allow_html=True)

            with components["metric_plot"].container():
                st.subheader("SampleSet Analysis")
                # line
                st.subheader("Metric")
                plot_components["line"] = st.empty()
                plot_components["settings_line"] = st.empty()

                with plot_components["settings_line"].container():
                    array_options: list[str] = [
                        c.split("_mean")[0]
                        for c in evaluation_table.filter(regex="_mean").columns
                    ]
                    metric_options = [
                        "success_probability",
                        "feasible_rate",
                        "residual_energy",
                        "TTS[optimal]",
                        "TTS[feasible]",
                        "TTS[derived]",
                    ]
                    groupby_options: list[str] = params_table.columns.tolist()

                    cols = st.columns(2)
                    with cols[0]:
                        groupby = st.multiselect(
                            "Group by:",
                            options=groupby_options,
                            default=groupby_options[0],
                            key="groupby",
                        )
                        x_labels = [o for o in groupby_options if o not in groupby]
                    with cols[1]:
                        base_y_labels = (
                            st.multiselect(
                                "Metrics:",
                                options=array_options + metric_options,
                                default=array_options[0],
                                key="metric",
                            )
                            or array_options[0]
                        )

                figs = []
                for base_y_label in base_y_labels:
                    if base_y_label in array_options:
                        if "violations" in base_y_label:
                            y_label = base_y_label + "_min"
                        else:
                            y_label = base_y_label + "_mean"
                    else:
                        y_label = base_y_label

                    fig = go.Figure()
                    for i, (name, group) in enumerate(params_table.groupby(groupby)):
                        if isinstance(name, str):
                            name = [name]
                        name = [str(n) for n in name]

                        line_color = px.colors.sequential.Jet[
                            i % len(px.colors.sequential.Jet)
                        ]
                        fill_color = px.colors.sequential.Jet[
                            i % len(px.colors.sequential.Jet)
                        ]

                        x = (
                            group[x_labels]
                            .astype(str)
                            .apply(lambda x: ", ".join(x), axis=1)
                        )
                        y = evaluation_table.loc[group.index, y_label]
                        upper = (
                            y + evaluation_table.loc[group.index, f"{base_y_label}_std"]
                        )
                        lower = (
                            y - evaluation_table.loc[group.index, f"{base_y_label}_std"]
                        )

                        if "violations" in base_y_label:
                            fig.add_traces(
                                [
                                    go.Scatter(
                                        x=x,
                                        y=y,
                                        name=", ".join(name),
                                        mode="markers+lines",
                                        line=dict(color=line_color),
                                    )
                                ]
                            )
                        else:
                            fig.add_traces(
                                [
                                    go.Scatter(
                                        x=x,
                                        y=y,
                                        name=", ".join(name),
                                        mode="markers+lines",
                                        line=dict(color=line_color),
                                    ),
                                    go.Scatter(
                                        x=x,
                                        y=upper,
                                        name=", ".join(name),
                                        showlegend=False,
                                        mode="lines",
                                        fillcolor=_rgb_to_rgba(fill_color, 0.2),
                                        line=dict(color=_rgb_to_rgba(fill_color, 0.0)),
                                    ),
                                    go.Scatter(
                                        x=x,
                                        y=lower,
                                        fill="tonexty",
                                        name=", ".join(name),
                                        showlegend=False,
                                        mode="lines",
                                        fillcolor=_rgb_to_rgba(fill_color, 0.2),
                                        line=dict(color=_rgb_to_rgba(fill_color, 0.0)),
                                    ),
                                ]
                            )
                        fig.update_layout(
                            title=f"{y_label.replace('_mean', '').capitalize()} grouped by {groupby}",
                            xaxis=dict(
                                automargin=True,
                                title=", ".join(x_labels),
                                type="category",
                                showgrid=True,
                            ),
                            yaxis=dict(
                                automargin=True, title=base_y_label, showgrid=True
                            ),
                            coloraxis=dict(colorscale="jet"),
                            legend=dict(x=1, y=1, xanchor="left", yanchor="top"),
                        )
                    figs.append(fig)

                with plot_components["line"].container():
                    for fig in figs:
                        st.plotly_chart(fig, use_container_width=True)

            # with components["scatter"].container():
            #     st.subheader("Scatter")
            #     cols = st.columns(3)
            #     with cols[0]:
            #         xlabel = st.selectbox("X:", options=options, index=0)
            #     with cols[1]:
            #         ylabel = st.selectbox("Y:", options=options, index=1)
            #     with cols[2]:
            #         color = st.selectbox("Color:", options=options, index=2)
            #     fig = px.scatter(results_table, x=xlabel, y=ylabel, color=color)
            #     fig.update_layout(margin=dict(t=10))
            #     st.plotly_chart(fig, use_container_width=True)
            #     st.markdown("<hr>", unsafe_allow_html=True)

            # parallel_coordinates
            with components["parallel_coordinates"].container():
                categorical_columns = params_table.select_dtypes(
                    include="object"
                ).columns.tolist()

                for name, _ in params_table.groupby(categorical_columns):
                    plot_components[
                        f"parallel_coordinates_{'_'.join(name)}"
                    ] = st.empty()

                metrics = ["feasible_rate"]
                filterd_evaluation_table = pd.concat(
                    [evaluation_table.filter(regex="_min"), evaluation_table[metrics]],
                    axis=1,
                ).rename(columns=lambda x: x.replace("_min", ""))
                cols = st.columns(4)
                with cols[0]:
                    color_options = filterd_evaluation_table.columns
                    color = st.selectbox(
                        "Color:",
                        options=color_options,
                        index=color_options.get_loc("objective"),
                    )

                for name, group in params_table.groupby(categorical_columns):
                    parcoords_table = pd.concat(
                        [
                            group,
                            filterd_evaluation_table.loc[group.index],
                        ],
                        axis=1,
                    )
                    max_column_length = max([len(str(c)) for c in parcoords_table])

                    fig = px.parallel_coordinates(
                        parcoords_table.reset_index(drop=True),
                        color=color,
                        color_continuous_scale="tealrose",
                    )
                    fig.update_traces(labelangle=30)
                    fig.update_coloraxes(colorbar=dict(titleside="right"))
                    fig.update_layout(
                        margin=dict(l=100, t=3.85 * max_column_length + 25),
                        title_text=", ".join(name),
                        title_y=0.98,
                    )

                    with plot_components[f"parallel_coordinates_{'_'.join(name)}"]:
                        st.plotly_chart(fig, use_container_width=True)

                # st.markdown("<hr>", unsafe_allow_html=True)

            with components["sampleset_analysis"].container():
                st.subheader("Constraint violations")

                # violations_analysis_cols = st.columns(2)

                # bar
                plot_components["bar"] = st.empty()
                plot_components["settings_bar"] = st.empty()
                # st.markdown("<hr>", unsafe_allow_html=True)
                with plot_components["settings_bar"]:
                    agg_options = ["min", "max", "mean", "std"]
                    groupby_options: list[str] = params_table.columns.tolist()
                    cols = st.columns(4)
                    with cols[0]:
                        agg = (
                            st.selectbox(
                                "Select a function for aggregating violations:",
                                options=agg_options,
                                index=0,
                            )
                            or "min"
                        )
                    with cols[1]:
                        groupby = (
                            st.selectbox("Group by:", options=groupby_options, index=0)
                            or groupby_options[0]
                        )
                table = pd.concat([params_table, evaluation_table], axis=1)
                unselected_options = [c for c in params_table if c != groupby]
                violations_table = table.filter(
                    regex="|".join(groupby_options + [f"violations_{agg}"])
                )
                violations_table = violations_table.rename(
                    columns={
                        c: c.replace(f"_violations_{agg}", "")
                        for c in violations_table.columns
                        if c.endswith(f"violations_{agg}")
                    }
                )
                violations_table = violations_table.astype(
                    {c: str for c in groupby_options}
                )
                stacked = violations_table.set_index(
                    params_table.columns.tolist()
                ).stack()
                stacked.name = f"violations_{agg}"
                stacked = stacked.reset_index(unselected_options)
                fig = px.bar(
                    stacked,
                    x=stacked.index.get_level_values(-1),
                    y=stacked[f"violations_{agg}"],
                    color=stacked.index.get_level_values(groupby),
                    barmode="group",
                    hover_data=unselected_options,
                    color_discrete_sequence=px.colors.sequential.Jet,
                )
                with plot_components["bar"].container():
                    st.plotly_chart(fig, use_container_width=True)

            constraint_heatmap = st.columns(2)
            # constraint violations
            with constraint_heatmap[0]:
                numeric_columns = params_table.select_dtypes("number").columns
                categorical_columns = params_table.select_dtypes("object").columns
                st.subheader("Heatmap for constraint violations")
                for name, _ in params_table.groupby(categorical_columns.tolist()):
                    plot_components[f"heatmap_violations_{'_'.join(name)}"] = st.empty()

                violations_table = evaluation_table.filter(regex="violations_min")

                cols = st.columns(2)
                with cols[0].container():
                    x_labels = st.multiselect(
                        "x:",
                        options=numeric_columns,
                        default=numeric_columns[0],
                        key="constraint_violations",
                    )
                for name, group in params_table.groupby(categorical_columns.tolist()):
                    fig = px.imshow(
                        violations_table.loc[group.index].T,
                        x=group[x_labels]
                        .astype(str)
                        .apply(lambda x: ", ".join(x), axis=1),
                        y=violations_table.loc[group.index]
                        .rename(columns=lambda x: x.replace("_violations_min", ""))
                        .columns,
                        labels=dict(x=f"{', '.join(x_labels)}", y="Constraint name"),
                        color_continuous_scale="deep_r",
                        aspect="auto",
                    )
                    fig.update_layout(
                        title_text=", ".join(name),
                        title_x=0,
                        xaxis=dict(automargin=True, type="category"),
                        yaxis=dict(automargin=True),
                    )
                    with plot_components[
                        f"heatmap_violations_{'_'.join(name)}"
                    ].container():
                        st.plotly_chart(fig, use_container_width=True)
                # st.markdown("<hr>", unsafe_allow_html=True)

            # constraint expr values
            with constraint_heatmap[1]:
                st.subheader("Heatmap for constraint expression values")
                for name, _ in params_table.groupby(categorical_columns.tolist()):
                    plot_components[
                        f"heatmap_expression_values_{'_'.join(name)}"
                    ] = st.empty()

                def expand(x: jm.SampleSet) -> pd.Series:
                    violations = x.evaluation.constraint_violations or {}
                    violations_argmin_map = {
                        k: np.argmin(v) for k, v in violations.items()
                    }

                    expr_values_map = {k: {} for k in violations}
                    for const_name in violations:
                        argmin = violations_argmin_map[const_name]
                        expr_values_map[const_name].update(
                            x.evaluation.constraint_expr_values[argmin][const_name]
                        )

                    return expr_values_map

                concat: jb.functions.Concat[jb.Experiment] = jb.functions.Concat()
                results = concat(
                    [
                        jb.load(benchmark_id, savedir=session.state.logdir)
                        for benchmark_id in benchmark_ids
                    ]
                )
                _, t = results.data
                sampleset_column = [
                    c for c in t.data if isinstance(t.data[c][0], jb.SampleSet)
                ][0]
                sampleset_table = t.data.filter([sampleset_column]).applymap(
                    lambda x: x.data
                )
                sampleset_table = pd.concat([params_table, sampleset_table], axis=1)

                sampleset_table = pd.concat(
                    [params_table, sampleset_table[sampleset_column].apply(expand)],
                    axis=1,
                )

                cols = st.columns(2)
                with cols[0].container():
                    const_name = st.selectbox(
                        "Select constraint:",
                        options=sampleset_table[sampleset_column][0].keys(),
                        index=0,
                    )
                with cols[1].container():
                    x_labels = st.multiselect(
                        "x:",
                        options=numeric_columns,
                        default=numeric_columns[0],
                        key="constraint_expr_values",
                    )

                for name, group in sampleset_table.groupby(
                    categorical_columns.tolist()
                ):
                    index = list(
                        map(
                            lambda x: str(x),
                            group[sampleset_column][0][const_name].keys(),
                        )
                    )
                    x = (
                        group[x_labels]
                        .astype(str)
                        .apply(lambda x: ", ".join(x), axis=1)
                    )
                    expr_values = np.array(
                        group[sampleset_column]
                        .apply(lambda x: list(x[const_name].values()))
                        .tolist(),
                        dtype=np.float64,
                    ).T

                    fig = px.imshow(
                        expr_values,
                        x=x,
                        y=index,
                        color_continuous_scale="deep_r",
                        labels=dict(x=f"{', '.join(x_labels)}", y="Index"),
                        aspect="auto",
                    )
                    fig.update_layout(
                        title_text=", ".join(name),
                        title_x=0,
                        xaxis=dict(automargin=True, type="category"),
                        yaxis=dict(automargin=True),
                        coloraxis=dict(
                            colorbar=dict(title=const_name, titleside="right")
                        ),
                    )

                    with plot_components[
                        f"heatmap_expression_values_{'_'.join(name)}"
                    ].container():
                        st.plotly_chart(fig, use_container_width=True)
                # st.markdown("<hr>", unsafe_allow_html=True)

            # tmp
            # concat: jb.functions.Concat[jb.Experiment] = jb.functions.Concat()
            # results = concat(
            #     [
            #         jb.load(benchmark_id, savedir=session.state.logdir)
            #         for benchmark_id in benchmark_ids
            #     ]
            # )
            # _, t = results.data
            # heatmap = []
            # for name, group in t.view().groupby("solver_name"):
            #     for sampleset in t.data.loc[group.index, "solver_output[0]"]:
            #         expr_values = sampleset.evaluation.constraint_expr_values[-1]
            #         expr_values = {str(k): v for k, v in expr_values.items()}
            #         x = expr_values["assign"]
            #         keys = [str(k) for k in x.keys()]
            #         values = list(x.values())
            #         heatmap.append(values)
            # heatmap = np.array(heatmap).T
            # fig = px.imshow(heatmap, aspect="auto", color_continuous_scale="Viridis")
            # st.plotly_chart(fig, use_container_width=True)
            # st.markdown("<hr>", unsafe_allow_html=True)

            # with ph_sampleset_diff.container():
            #    st.subheader("Diff")
            #    cols = st.columns(2)
            #    with cols[0]:
            #        r1_name = st.selectbox(
            #            "Record 1", options=range(len(results_table)), index=0
            #        )
            #    with cols[1]:
            #        r2_name = st.selectbox(
            #            "Record 2", options=range(len(results_table)), index=1
            #        )

            #    with elements("diff"):
            #        results = jb.load(benchmark_id, savedir=session.state.logdir)
            #        r1 = results.data[1].data.iloc[r1_name]["solver_output[0]"]
            #        r2 = results.data[1].data.iloc[r2_name]["solver_output[0]"]
            #        editor.MonacoDiff(
            #            original="\n".join(
            #                r1.__repr__()[i : i + 100]
            #                for i in range(0, len(r1.__repr__()), 100)
            #            ),
            #            modified="\n".join(
            #                r2.__repr__()[i : i + 100]
            #                for i in range(0, len(r2.__repr__()), 100)
            #            ),
            #            height=300,
            #        )


@st.cache_data
def _get_params_table(benchmark_ids: list[str], savedir: pathlib.Path) -> pd.DataFrame:
    table = jb.load(benchmark_ids[0], savedir=savedir).params_table
    expected_problem_columns = [c for c in table if isinstance(table[c][0], jm.Problem)]
    for benchmark_id in benchmark_ids[1:]:
        params_table = jb.load(benchmark_id, savedir=savedir).params_table
        problem_columns = [
            c for c in params_table if isinstance(params_table[c][0], jm.Problem)
        ]
        for c, expected in zip(problem_columns, expected_problem_columns):
            params_table = params_table.rename(columns={c: expected})

        table = pd.concat([table, params_table])

    for c in ("feed_dict", "instance_data"):
        droped_columns = table.filter(regex=c).columns
        table = table.drop(columns=droped_columns)

    for c in expected_problem_columns:
        table[c] = table[c].apply(lambda x: x.name)
    return table


@st.cache_data
def _get_problem_table(benchmark_ids: list[str], savedir: pathlib.Path) -> pd.DataFrame:
    table = jb.load(benchmark_ids[0], savedir=savedir).params_table
    expected_problem_columns = [c for c in table if isinstance(table[c][0], jm.Problem)]
    for benchmark_id in benchmark_ids[1:]:
        params_table = jb.load(benchmark_id, savedir=savedir).params_table
        problem_columns = [
            c for c in params_table if isinstance(params_table[c][0], jm.Problem)
        ]
        for c, expected in zip(problem_columns, expected_problem_columns):
            params_table = params_table.rename(columns={c: expected})

        table = pd.concat([table, params_table])
    return table[expected_problem_columns]


@st.cache_data
def _get_response_table(
    benchmark_ids: list[str], savedir: pathlib.Path
) -> pd.DataFrame:
    table = pd.DataFrame()
    for benchmark_id in benchmark_ids:
        results = jb.load(benchmark_id, savedir=savedir)
        response_table = jb.Table._expand_dict_in(results.response_table)
        table = pd.concat([table, response_table])
    return table


@st.cache_data
def _get_evaluation_table(
    benchmark_ids: list[str],
    savedir: pathlib.Path,
    opt_value: float | None = None,
    pr: float = 0.99,
) -> pd.DataFrame:
    table = pd.DataFrame()
    for benchmark_id in benchmark_ids:
        results = jb.load(benchmark_id, savedir=savedir)
        e = jb.Evaluation()
        evaluation_table = e([results], opt_value=opt_value, pr=pr).table.drop(
            columns=results.table.columns,
        )
        table = pd.concat([table, evaluation_table])
    return table


def _rgb_to_rgba(rgb_str: str, alpha: float) -> str:
    rgba = rgb_str.replace("rgb", "rgba")
    rgba = rgba.split(")")
    rgba[-1] = f"{alpha}"
    return ",".join(rgba) + ")"


def _is_number_str(s: str) -> bool:
    pattern = r"^[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?$"
    return bool(re.match(pattern, s))



from __future__ import annotations

import inspect
import pathlib
import typing as tp
from functools import wraps

import jijbench as jb
from jijbench.consts.default import DEFAULT_RESULT_DIR


def checkpoint(
    name: str | None = None, savedir: str | pathlib.Path = DEFAULT_RESULT_DIR
):
    """Decorator for saving and checkpointing the results of a function.

    Args:
        name (str, optional): The name of the benchmark. Defaults to None.
        savedir (str | pathlib.Path, optional): The directory where the benchmark will be saved. Defaults to DEFAULT_RESULT_DIR.

    Returns:
        Callable: The decorated function.

    Examples:
        ```python
        # case POSITIONAL OR KEYWORD
        @checkpoint(name="example_checkpoint")
        def f1(cha: str, i: int = 1) -> str:
            return f"{cha}-{i}"

        # case KEYWORD_ONLY
        @checkpoint(name="example_checkpoint")
        def f2(cha: str, *, f: float=1.0) -> str:
            return f"{cha}-{f}"

        # case VAR_KEYWORD
        # Variable positional arguments like *args are not supported.
        @checkpoint(name="example_checkpoint")
        def f3(cha: str, **kwargs) -> str:
            return f"{cha}-{kwargs}"

        f1("1", 2)
        f2("1", f=2.0)
        f3("1", f=2.0, g=3.0)

        result = jb.load("example_checkpoint")
        ```
    """

    def decorator(func: tp.Callable[..., tp.Any]):
        @wraps(func)
        def wrapper(*args: tp.Any, **kwargs: tp.Any):
            params = {}
            signature = inspect.signature(func)
            pos_arg_index = 0
            kw_arg_index = 0

            # Make benchmark parameters by inspecting the function signature.
            # k is argument name, v is inspect.Parameter
            # v.kind = 1 means POSITIONAL OR KEYWORD. ex. def f(a, b=1)
            # v.kind = 2 means VAR_POSITIONAL. ex. def f(*args)
            # v.kind = 3 means KEYWORD_ONLY. ex. def f(*, a)
            # v.kind = 4 means VAR_KEYWORD. ex. def f(**kwargs)
            for k, v in signature.parameters.items():
                if v.kind == 1:
                    if pos_arg_index < len(args):
                        params[k] = [args[pos_arg_index]]
                        pos_arg_index += 1
                elif v.kind == 2:
                    raise TypeError("Variable positional arguments are not supported.")
                elif v.kind == 3:
                    if k in kwargs:
                        params[k] = [kwargs[k]]
                        kw_arg_index += 1
                elif v.kind == 4:
                    while kw_arg_index < len(kwargs):
                        k = list(kwargs.keys())[kw_arg_index]
                        if k in kwargs:
                            params[k] = [kwargs[k]]
                            kw_arg_index += 1
            bench = jb.Benchmark(solver=func, params=params, name=name)
            if name:
                bench_dir = pathlib.Path(savedir) / name
                if bench_dir.exists():
                    jb.load(bench_dir).apply(bench, savedir=savedir)
                else:
                    bench(savedir=savedir)
            else:
                bench(savedir=savedir)
            return func(*args, **kwargs)

        return wrapper

    return decorator



from __future__ import annotations

import jijmodeling as jm


def travelling_salesman():
    # 問題
    problem = jm.Problem("travelling-salesman")
    dist = jm.Placeholder("d", dim=2)
    N = jm.Placeholder("N")

    x = jm.Binary("x", shape=(N, N))
    i = jm.Element("i", N)
    j = jm.Element("j", N)

    t = jm.Element("t", N)

    # Objective Funtion
    sum_list = [t, i, j]
    obj = jm.Sum(sum_list, dist[i, j] * x[t, i] * x[(t + 1) % N, j])
    problem += obj

    # const1: onehot for time
    const1 = x[t, :]
    problem += jm.Constraint(
        "onehot-time",
        const1 == 1,
        forall=[
            t,
        ],
    )

    # const2: onehot for location
    const2 = x[:, i]
    problem += jm.Constraint(
        "onehot-location",
        const2 == 1,
        forall=[
            i,
        ],
    )

    return problem



from __future__ import annotations

import typing as tp
from dataclasses import dataclass

import numpy as np

from jijbench.elements.base import Number
from jijbench.functions.math import Max, Mean, Min, Std
from jijbench.node.base import DataNode
from jijbench.typing import ArrayType


@dataclass
class Array(DataNode[ArrayType]):
    """A class representing numpy arrays.

    Attributes:
        data (numpy.ndarray): The numpy array.
        name (str): The name of the node.
    """

    def min(self) -> Number:
        """Get the minimum value of the numpy array.

        Returns:
            Number: The minimum value of the numpy array.
        """
        return self.apply(Min())

    def max(self) -> Number:
        """Get the maximum value of the numpy array.

        Returns:
            Number: The maximum value of the numpy array.
        """
        return self.apply(Max())

    def mean(self) -> Number:
        """Get the mean value of the numpy array.

        Returns:
            Number: The mean value of the numpy array.
        """
        return self.apply(Mean())

    def std(self) -> Number:
        """Get the standard deviation of the numpy array.

        Returns:
            Number: The standard deviation of the numpy array.
        """
        return self.apply(Std())

    @classmethod
    def validate_data(cls, data: ArrayType) -> ArrayType:
        """Validate the data to ensure it is a numpy array.

        Args:
            data (numpyp.ndarray): The numpy array to validate.

        Raises:
            TypeError: If the input data is not a numpy array.

        Returns:
            numpy.ndarray: The validated numpy array.
        """
        return cls._validate_dtype(data, (np.ndarray,))



from __future__ import annotations

import jijmodeling as jm


def travelling_salesman_with_time_windows():
    # 問題
    problem = jm.Problem("travelling-salesman-with-time-windows")

    # 距離行列
    dist = jm.Placeholder("d", dim=2)  # 距離行列
    N = jm.Placeholder("N")
    e = jm.Placeholder("e", shape=(N,))  # ready time
    l = jm.Placeholder("l", shape=(N,))  # due time
    x = jm.Binary("x", shape=(N, N))
    t = jm.Integer("t", shape=(N,), lower=e, upper=l)

    i = jm.Element("i", N)
    j = jm.Element("j", N)

    # Objevtive Function: 距離の最小化
    sum_list = [i, (j, j != i)]
    obj = jm.Sum(sum_list, dist[i, j] * x[i, j])
    problem += obj

    # Const1: 都市iから出る辺は1つ
    term1 = jm.Sum((j, j != i), x[i, j])
    problem += jm.Constraint(
        "onehot-constraint1",
        term1 == 1,
        forall=[
            i,
        ],
    )

    # Const2: 都市iに入る辺は1つ
    term2 = jm.Sum((j, j != i), x[j, i])
    problem += jm.Constraint(
        "onehot-constraint2",
        term2 == 1,
        forall=[
            i,
        ],
    )

    # Const3: Time Windows制約
    term3 = t[i] + dist[i, j] - t[j] - 20 * (1 - x[i, j])
    forall_list = [(j, j != 0), (i, (i != 0) & (i != j))]
    problem += jm.Constraint("time-window-constraint", term3 <= 0, forall=forall_list)

    return problem



from jijbench.datasets.problems.bin_packing import bin_packing
from jijbench.datasets.problems.knapsack import knapsack
from jijbench.datasets.problems.nurse_scheduling import nurse_scheduling
from jijbench.datasets.problems.travelling_salesman import travelling_salesman
from jijbench.datasets.problems.travelling_salesman_with_time_windows import (
    travelling_salesman_with_time_windows,
)

__all__ = [
    "bin_packing",
    "knapsack",
    "nurse_scheduling",
    "travelling_salesman",
    "travelling_salesman_with_time_windows",
]



from __future__ import annotations

from typing import Callable, cast

import matplotlib.pyplot as plt
from matplotlib.ticker import MaxNLocator

import jijbench as jb
from jijbench.visualization.metrics.utils import _create_fig_title_list


class BasePlot:
    def __init__(self, result: jb.Experiment) -> None:
        """Visualize the metrics of a benchmark result by boxplot.

        Attributes:
            result (jb.Experiment): a benchmark result.
        """
        self.result = result

    def boxplot(
        self,
        f: Callable,
        figsize: tuple[int | float] | None = None,
        title: str | list[str] | None = None,
        title_fontsize: float | None = None,
        xticklabels_size: float | None = None,
        xticklabels_rotation: float | None = None,
        ylabel: str | None = None,
        ylabel_size: float | None = None,
        yticks: list[int | float] | None = None,
        **matplotlib_boxplot_kwargs,
    ):
        """Draw a box and whisker plot of the metrics based on `result` data using matplotlib.boxplot.

        This method applies the function f to the result (i.e. `jb.Experiment`) to get the metrics (pd.Series), and draw boxplot of this metrics.
        This metrics series calculated as metrics = self.result.table.apply(f, axis=1) in this method assumes the following structure.
            The length is equal to the number of rows in `result.table`.
            the element is a dictionary where the key is the name of each boxplot and the value is the np.array of the boxplot data.
        This method returns a figure and axes, so you can post-process them to change the appearance of the plot.
        See also the example below.

        Args:
            f (Callable): Callalbe to apply to table and get metrics.
            figsize (tuple[int | float] | None): the size of figure. The default uses matplotlib's default value.
            title (str | list[str] | None): the title of figure. The default uses the indices of `result.table`.
            title_fontsize (float | None): the fontsize of the title.The default uses matplotlib's default value.
            xticklabels_size (float | None): the fontsize of the xticklabels (i.e. the name of each boxplot). The default uses matplotlib's default value.
            xticklabels_rotation (float | None): the rotation angle of the xticklabels in degree.The default uses matplotlib's default value.
            ylabel (str | None): the ylabel of figure. Defaults to None.
            ylabel_size (float | None): the fontsize of the ylabel. The default uses matplotlib's default value.
            yticks (list[int | float] | None): the yticks of figure. Default to only integers by`MaxNLocator(integer=True)`.
            **matplotlib_boxplot_kwargs (dict): the parameter passed to matplotlib.boxplot.

        Returns:
            tuple[tuple[matplotlib.figure.Figure, matplotlib.axes.Subplot]]: A tuple of length equal to the number of rows in result. each element of is a tuple of figure and axes.

        Example:
            The code below draws a boxplot of violations of each constraint.
            Note that result (i.e. `jb.Experiment`) holds violations for each constraint in np.array format.
            In the first example, postprocessing the figure and axes changes the appearance of the plot.

            ```python
            import jijbench as jb
            from jijbench.visualization import  BasePlot
            import jijzept as jz
            from jijzept.sampler.openjij.sa_cpu import JijSAParameters
            import pandas as pd

            problem = jb.get_problem("TSP")
            instance_data = jb.get_instance_data("TSP")[0][1]
            multipliers1 = {"onehot_time": 0.003, "onehot_location": 0.003}
            multipliers2 = {"onehot_time": 0.3, "onehot_location": 0.3}

            config_path = "XX"
            sa_parameter = JijSAParameters(num_reads=15)
            sa_sampler = jz.JijSASampler(config=config_path)

            bench = jb.Benchmark(
                params = {
                    "model": [problem],
                    "feed_dict": [instance_data],
                    "parameters": [sa_parameter],
                    "multipliers": [multipliers1, multipliers2],
                },
                solver = [sa_sampler.sample_model],
            )
            result = bench()

            def get_violations_dict(x: pd.Series) -> dict:
                constraint_violations_indices = x.index[x.index.str.contains("violations")]
                return {index: x[index] for index in constraint_violations_indices}

            bplot = BasePlot(result)
            fig_ax_tuple = bplot.boxplot(f=get_violations_dict)

            # you can post-process figure and axes to change the appearance of the plot.
            for fig, ax in fig_ax_tuple:
                fig.suptitle("my title")
                display(fig)
            ```

            By using the `construct_experiment_from_samplesets function`,
            `boxplot` can also be used for `jm.SampleSet` obtained without `JijBenchmark`.

            ```python
            import jijbench as jb
            import jijzept as jz
            from jijbench.visualization import  BasePlot
            from jijbench.visualization import construct_experiment_from_samplesets
            import pandas as pd

            problem = jb.get_problem("TSP")
            instance_data = jb.get_instance_data("TSP")[0][1]
            multipliers = {"onehot_time": 0.003, "onehot_location": 0.003}

            config_path = "XX"
            sampler = jz.JijSASampler(config=config_path)
            sampleset = sampler.sample_model(model=problem, feed_dict=instance_data, multipliers=multipliers, num_reads=100)

            def get_violations_dict(x: pd.Series) -> dict:
                constraint_violations_indices = x.index[x.index.str.contains("violations")]
                return {index: x[index] for index in constraint_violations_indices}

            result = construct_experiment_from_samplesets(sampleset)
            bplot = BasePlot(result)
            fig_ax_tuple = bplot.boxplot(f=get_violations_dict)
            ```
        """
        metrics = self.result.table.apply(f, axis=1)
        title_list = _create_fig_title_list(metrics, title)

        fig_ax_list = []
        for i, (_, data) in enumerate(metrics.items()):
            fig, ax = plt.subplots(figsize=figsize)
            fig.suptitle(title_list[i], fontsize=title_fontsize)
            ax.boxplot(data.values(), **matplotlib_boxplot_kwargs)
            ax.set_xticklabels(
                data.keys(), size=xticklabels_size, rotation=xticklabels_rotation
            )
            ylabel = cast("str", ylabel)
            ax.set_ylabel(ylabel, size=ylabel_size)
            if yticks is None:
                # make yticks integer only
                ax.yaxis.set_major_locator(MaxNLocator(integer=True))
            else:
                ax.set_yticks(yticks)
            fig_ax_list.append((fig, ax))

        return tuple(fig_ax_list)



from pkgutil import extend_path

__path__ = extend_path(__path__, __name__)

from jijbench.visualization.figure.graph import Graph, GraphType
from jijbench.visualization.figure.schedule import Schedule
from jijbench.visualization.figure.timeseries import TimeSeries
from jijbench.visualization.metrics.baseplot.baseplot import BasePlot
from jijbench.visualization.metrics.constraintplot.constraintplot import ConstraintPlot
from jijbench.visualization.metrics.parallelplot.parallelplot import MetricsParallelPlot
from jijbench.visualization.metrics.utils import construct_experiment_from_samplesets

__all__ = [
    "BasePlot",
    "ConstraintPlot",
    "construct_experiment_from_samplesets",
    "MetricsParallelPlot",
    "Graph",
    "GraphType",
    "Schedule",
    "TimeSeries",
]



from __future__ import annotations

import jijmodeling as jm


def nurse_scheduling():
    # 問題
    problem = jm.Problem("nurse-scheduling")

    I = jm.Placeholder("I")  # 人の数
    D = jm.Placeholder("D")  # 日数 D%7=0 とし, 月曜日からstartするとする
    W = jm.Placeholder("W")  # 週の数
    T = jm.Placeholder("T")  # シフトのタイプ数

    N = jm.Placeholder("N", dim=2)  # N[i] = 人iが必ず休む日

    # シフトtの次の日に働ないtypeのシフト R = [type1, type2, type3, ... ] = [[1, 2], [0, ], [], ...]
    R = jm.Placeholder("R", dim=2)
    l = jm.Placeholder("l", shape=(I,))  # シフトtの労働時間
    m_max = jm.Placeholder("m_max", shape=(I, T)).set_latex(
        r"\mathrm{m\_max}"
    )  # 従業員にシフトtを割当てられる最大回数
    b_min = jm.Placeholder("b_min", shape=(I,)).set_latex(
        r"\mathrm{b\_min}"
    )  # 各従業員の労働時間の最小値
    b_max = jm.Placeholder("b_max", shape=(I,)).set_latex(
        r"\mathrm{b\_max}"
    )  # 各従業員の労働時間の最大値
    c_min = jm.Placeholder("c_min", shape=(I,)).set_latex(
        r"\mathrm{c\_min}"
    )  # 各従業員の最小連続勤務数
    c_max = jm.Placeholder("c_max", shape=(I,)).set_latex(
        r"\mathrm{c\_max}"
    )  # 各従業員の最大連続勤務数
    o_min = jm.Placeholder("o_min", shape=(I,)).set_latex(
        r"\mathrm{o\_min}"
    )  # 各従業員の連続最低休日日数
    a_max = jm.Placeholder("a_max", shape=(I,)).set_latex(
        r"\mathrm{a\_max}"
    )  # 各従業員の週末働ける最大の回数
    # q[i, d, t] = 1, 人iは日にちdにtype tの仕事をしたい
    q = jm.Placeholder("q", shape=(I, D, T))
    # p[i, d, t] = 1, 人iは日にちdにtype tの仕事をしたくない
    p = jm.Placeholder("p", shape=(I, D, T))
    u = jm.Placeholder("u", shape=(D, T))  # 日にちd, type tの必要人数
    v_min = jm.Placeholder("v_min", shape=(D, T)).set_latex(
        r"\mathrm{v\_min}"
    )  # 人員不足のペナルティーの重み
    v_max = jm.Placeholder("v_max", shape=(D, T)).set_latex(
        r"\mathrm{v\_max}"
    )  # 人員過剰のペナルティーの重み

    len_R = R.shape[0]

    # Element
    i = jm.Element("i", (0, I))
    d = jm.Element("d", (0, D))
    t = jm.Element("t", (0, T))
    w = jm.Element("w", (0, W))
    dt = jm.Element("dt", (0, len_R))
    j = jm.Element("j", (0, D))
    s = jm.Element("s", (0, D))

    # 決定変数
    x = jm.Binary("x", shape=(I, D, T))  # 人iを日にちdにシフトtを割当てるかどうか
    k = jm.Binary("k", shape=(I, W))  # week wに働いたかどうか
    y = jm.Integer("y", lower=0, upper=u, shape=(D, T))  # 日dのシフトtの不足人数
    z = jm.Integer("z", lower=0, upper=u, shape=(D, T))  # 日dのシフトtの過剰人数

    # Objective Function
    term1 = jm.Sum([i, d, t], q[i, d, t] * (1 - x[i, d, t]))  # シフト入りの希望を叶える
    term2 = jm.Sum([i, d, t], p[i, d, t] * x[i, d, t])  # シフト休みの希望を叶える
    term3 = jm.Sum([d, t], y[d, t] * v_min[d, t])  # 人員不足のペナルティー
    term4 = jm.Sum([d, t], z[d, t] * v_max[d, t])  # 人員過剰のペナルティー
    problem += term1 + term2 + term3 + term4

    # Constraint1: 1人一つの仕事しか割り当てられない
    const1 = jm.Sum(t, x[i, d, t])
    problem += jm.Constraint("assign", const1 - 1 <= 0, forall=[i, d])

    # Constraint2: 特定の仕事を行った次の日に, 特定の仕事を行うことはできない
    problem += jm.Constraint(
        "shift-rotation",
        x[i, d, t] + x[i, (d + 1), dt] - 1 <= 0,
        forall=[i, (d, (d != D - 1)), t, (dt, (R[t, dt] != -1))],
    )

    # Constraint3: 従業員に割り当てられる各タイプのシフトの最大数
    const3 = jm.Sum(d, x[i, d, t]) - m_max[i, t]
    problem += jm.Constraint("assign-max", const3 <= 0, forall=[i, t])

    # Constraint4: 最低・最高労働時間
    const4 = jm.Sum([d, t], l[t] * x[i, d, t])
    problem += jm.Constraint(
        "minimum-work-time",
        b_min[i] - const4 <= 0,
        forall=[
            i,
        ],
    )
    problem += jm.Constraint(
        "maximum-work-time",
        const4 - b_max[i] <= 0,
        forall=[
            i,
        ],
    )

    # Constraint5: 最大連続勤務
    j5 = jm.Element("j5", (d, d + c_max[i] + 1))
    const5 = jm.Sum([j5, t], x[i, j5, t])
    problem += jm.Constraint(
        "maximum-consecutive-shifts",
        const5 - c_max[i] <= 0,
        forall=[i, (d, d <= D - (c_max[i] + 1))],
    )

    # Constraint6: 最小連続勤務
    s6 = jm.Element("s6", (1, c_min[i]))
    j6 = jm.Element("j6", (d + 1, d + s6 + 1))
    problem += jm.Constraint(
        "minimum-consecutive-shifts",
        jm.Sum(t, x[i, d, t])
        + (s6 - jm.Sum([j6, t], x[i, j6, t]))
        + jm.Sum(t, x[i, d + s6 + 1, t])
        - 1
        >= 0,
        forall=[i, s6, (d, d < (D - (s6 + 1)))],
    )

    # Constraint7: 最低連続休暇日数
    s7 = jm.Element("s7", (1, o_min[i]))
    j7 = jm.Element("j7", (d + 1, d + s7 + 1))
    problem += jm.Constraint(
        "minimum-consecutive-days-off",
        (1 - jm.Sum(t, x[i, d, t]))
        + jm.Sum([j7, t], x[i, j7, t])
        + (1 - jm.Sum(t, x[i, d + s7 + 1, t]))
        - 1
        >= 0,
        forall=[i, s7, (d, d < (D - (s7 + 1)))],
    )

    # Constraint8: 週末休みの最大回数
    const8 = jm.Sum(t, x[i, 7 * (w + 1) - 2, t]) + jm.Sum(t, x[i, 7 * (w + 1) - 1, t])
    problem += jm.Constraint("variable-k-lower", k[i, w] - const8 <= 0, forall=[i, w])
    problem += jm.Constraint(
        "variable-k-upper", const8 - 2 * k[i, w] <= 0, forall=[i, w]
    )
    problem += jm.Constraint(
        "maximum-number-of-weekends",
        jm.Sum(w, k[i, w]) - a_max[i] <= 0,
        forall=[
            i,
        ],
    )

    # Constraint9: 働けない日の制約
    d_o = jm.Element("do", N[i]).set_latex(r"{\mathrm d\_o}")
    problem += jm.Constraint(
        "days-off", x[i, d_o, t] == 0, forall=[i, (d_o, d_o != -1), t]
    )

    # Constraint10: 必要人数に関する制約
    const10 = jm.Sum(i, x[i, d, t]) - z[d, t] + y[d, t] - u[d, t]
    problem += jm.Constraint("cover-requirements", const10 == 0, forall=[d, t])

    return problem



from __future__ import annotations

import os
import pathlib
import subprocess

import click

from jijbench.consts.default import DEFAULT_RESULT_DIR


@click.command()
@click.option(
    "--logdir", default=DEFAULT_RESULT_DIR, help="Directory to save benchmark results"
)
def main(logdir: str):
    os.environ["logdir"] = logdir
    os.environ["STREAMLIT_THEME_BACKGROUND_COLOR"] = "white"

    command = ["streamlit", "run", f"{pathlib.Path(__file__).parent}/app.py"]
    subprocess.run(command)


if __name__ == "__main__":
    main()



from __future__ import annotations

import pathlib
import typing as tp

import numpy as np
import plotly.graph_objects as go
import streamlit as st
from plotly.subplots import make_subplots

from jijbench.consts.default import DEFAULT_PROBLEM_NAMES
from jijbench.datasets.model import InstanceDataFileStorage

if tp.TYPE_CHECKING:
    from jijbench.dashboard.session import Session


class InstanceDataDirTree:
    """
    A class representing a directory tree of instance data files.

    Attributes:
        num_files_to_display (int): The number of files to display.
        node_map (dict[str, tp.Any]): A property to get or set the node map.
        nodes (list[dict[str, tp.Any]]): A property to get the list of nodes.
        problem_names (list[str]): A property to get or set the list of problem names.
    """

    def __init__(self) -> None:
        self._node_map: dict[str, tp.Any] = {}
        self._num_files_to_display = 5
        self._problem_names = DEFAULT_PROBLEM_NAMES

        for problem_name in DEFAULT_PROBLEM_NAMES:
            storage = InstanceDataFileStorage(problem_name)

            node: dict[str, tp.Any] = {
                "label": problem_name,
                "value": problem_name,
                "children": [],
            }
            for size in ["small", "medium", "large"]:
                files = storage.get_files(size, self.num_files_to_display)
                node["children"] += [
                    {
                        "label": size,
                        "value": f"{problem_name}&{size}",
                        "children": [
                            {
                                "label": f"sample-{i + 1:03}",
                                "value": file,
                            }
                            for i, file in enumerate(files[: self.num_files_to_display])
                        ],
                    },
                ]
            self._node_map[problem_name] = node

    @property
    def node_map(self) -> dict[str, tp.Any]:
        return self._node_map

    @node_map.setter
    def node_map(self, node_map: dict[str, tp.Any]) -> None:
        self._node_map = node_map

    @property
    def nodes(self) -> list[dict[str, tp.Any]]:
        return list(self._node_map.values())

    @property
    def problem_names(self) -> list[str]:
        return self._problem_names

    @problem_names.setter
    def problem_names(self, problem_names: list[str]) -> None:
        self._problem_names = problem_names

    @property
    def num_files_to_display(self) -> int:
        return self._num_files_to_display

    @num_files_to_display.setter
    def num_files_to_display(self, num_files_to_display: int) -> None:
        self._num_files_to_display = num_files_to_display


class InstanceDataHandler:
    """
    A class to handle instance data files, manage their loading, and visualize them using various plot types.

    """

    def on_add(self, session: Session) -> None:
        """Adds a new instance data file to the directory tree.

        Args:
            session (Session): The current session.
        """
        problem_name = session.state.input_problem_name
        instance_data_name = session.state.uploaded_instance_data_name
        logdir = session.state.logdir
        instance_data_dir_tree = session.state.instance_data_dir_tree

        file = f"{logdir}/{problem_name}/{instance_data_name}"
        if problem_name in instance_data_dir_tree.node_map:
            if file not in [
                child["value"]
                for child in instance_data_dir_tree.node_map[problem_name]["children"]
            ]:
                instance_data_dir_tree.node_map[problem_name]["children"] += [
                    {"label": instance_data_name, "value": file}
                ]
        else:
            instance_data_dir_tree.node_map[problem_name] = {
                "label": problem_name,
                "value": problem_name,
                "children": [{"label": instance_data_name, "value": file}],
            }

        instance_data_dir_tree.problem_names += [problem_name]

    def on_load(self, session: Session) -> None:
        """Loads selected instance data files and displays the chosen plot type.

        Args:
            session (Session): The current session.
        """
        files = session.state.selected_instance_data_files
        data_map = _load_instance_data(files)

        name = session.state.selected_instance_data_name
        fig_type = session.state.selected_figure_for_instance_data
        if name:
            key = name.split("/")[-1]
            if fig_type == "Histogram":
                self.on_select_histogram(data_map[key])
            elif fig_type == "Box":
                self.on_select_box(data_map[key])
            elif fig_type == "Violin":
                self.on_select_violin(data_map[key])

    def on_select_histogram(self, data: dict[str, list[int | float]]) -> None:
        """Creates a histogram plot of the given data.

        Args:
            data (dict[str, list[int | float]]): The instance data to be visualized.
        """
        fig = _plot_histogram_for_instance_data(data)
        with st.container():
            st.plotly_chart(fig, use_container_width=True)

    def on_select_box(self, data: dict[str, list[int | float]]) -> None:
        """Creates a box plot of the given data.

        Args:
            data (dict[str, list[int | float]]): The instance data to be visualized.
        """
        fig = _plot_box_for_instance_data(data)
        with st.container():
            st.plotly_chart(fig, use_container_width=True)

    def on_select_violin(self, data: dict[str, list[int | float]]) -> None:
        """Display a violin plot for the given instance data.


        Args:
            data (dict[str, list[int | float]]): The instance data to be visualized.
        """
        fig = _plot_violin_for_instance_data(data)
        with st.container():
            st.plotly_chart(fig, use_container_width=True)


@st.cache_data
def _load_instance_data(files: list[str]) -> dict[str, dict[str, list[int | float]]]:
    def _flatten(data: int | float | list[tp.Any]) -> list[tp.Any]:
        if isinstance(data, list):
            return [xij for xi in data for xij in _flatten(xi)]
        else:
            return [data]

    data = {}
    for file in files:
        ins_d = InstanceDataFileStorage.load(file)
        data[pathlib.Path(file).name] = {k: _flatten(v) for k, v in ins_d.items()}
    return data


def _plot_histogram_for_instance_data(data: dict[str, list[int | float]]) -> go.Figure:
    fig = make_subplots(rows=len(data), cols=1)
    for i, (k, v) in enumerate(data.items()):
        v = np.array(v)
        if len(np.unique(v)) == 1:
            xmin, xmax = v[0] - 10, v[0] + 10
        else:
            xmin, xmax = v.min(), v.max()
        fig.add_trace(go.Histogram(x=v, name=k, nbinsx=100), row=i + 1, col=1)
        fig.update_xaxes(range=[xmin, xmax], row=i + 1, col=1)
    fig.update_layout(
        margin=dict(l=10, r=10, t=50, b=10),
        height=200 * len(data),
        width=700,
    )
    return fig


def _plot_box_for_instance_data(data: dict[str, list[int | float]]) -> go.Figure:
    fig = make_subplots(rows=1, cols=len(data))
    for i, (k, v) in enumerate(data.items()):
        fig.add_trace(go.Box(y=v, name=k), row=1, col=i + 1)
    fig.update_layout(margin=dict(l=10, r=10, t=50, b=10))
    return fig


def _plot_violin_for_instance_data(data: dict[str, list[int | float]]) -> go.Figure:
    fig = make_subplots(rows=1, cols=len(data))
    for i, (k, v) in enumerate(data.items()):
        fig.add_trace(
            go.Violin(y=v, x0=k, name=k, box_visible=True, meanline_visible=True),
            row=1,
            col=i + 1,
        )
    fig.update_layout(margin=dict(l=10, r=10, t=50, b=10))
    return fig



"""
This module defines the Session class, a singleton class responsible for managing the state of the dashboard,
as well as the State class, which has the state data of the application.

Classes:
Session: A Session class maintains the state of the application and the handlers responsible for different sections of the dashboard.
State: A State class holds session state attributes and provides properties to access and modify them.

Usage:
The Session class is used by the main dashboard application script to manage the state and interactions between different components of the dashboard.
"""

from __future__ import annotations

import pathlib
import typing as tp

import streamlit as st

from jijbench.consts.default import DEFAULT_RESULT_DIR
from jijbench.dashboard.handlers.instance_data import (
    InstanceDataDirTree,
    InstanceDataHandler,
)
from jijbench.dashboard.handlers.routing import RoutingHandler


class _Singleton(type):
    def __call__(cls, *args: tp.Any, **kwargs: tp.Any):
        if "_instance" not in st.session_state:
            st.session_state["_instance"] = super().__call__(*args, **kwargs)
        return st.session_state["_instance"]


class Session(metaclass=_Singleton):
    def __init__(self, logdir: pathlib.Path = DEFAULT_RESULT_DIR) -> None:
        self.state = State(logdir)
        self.routing_handler = RoutingHandler()
        self.instance_data_handler = InstanceDataHandler()

    def display_page(
        self, tab_name: tp.Literal["Instance data", "Problem", "Solver", "Analysis"]
    ) -> None:
        self.state.selected_page = tab_name
        self.routing_handler.on_select_page(self)

    def add_instance_data(self) -> None:
        self.instance_data_handler.on_add(self)

    def plot_instance_data(self) -> None:
        self.instance_data_handler.on_load(self)


class State:
    def __init__(self, logdir: pathlib.Path = DEFAULT_RESULT_DIR) -> None:
        st.session_state["selected_page"] = "Analysis"
        st.session_state["instance_data_dir_tree"] = InstanceDataDirTree()
        st.session_state["logdir"] = logdir
        st.session_state["input_problem_name"] = ""
        st.session_state["uploaded_instance_data_name"] = ""
        st.session_state["selected_instance_data_name"] = None
        st.session_state["is_instance_data_loaded"] = False
        st.session_state["selected_instance_data_map"] = {}
        st.session_state["selected_figure_for_instance_data"] = None
        st.session_state["selected_benchmark_results"] = []
        st.session_state["num_experiment_loaded"] = 0

    @property
    def selected_page(self) -> str:
        return st.session_state["selected_page"]

    @selected_page.setter
    def selected_page(self, page: str) -> None:
        st.session_state["selected_page"] = page

    @property
    def instance_data_dir_tree(self) -> InstanceDataDirTree:
        return st.session_state["instance_data_dir_tree"]

    @property
    def logdir(self) -> pathlib.Path:
        return st.session_state["logdir"]

    @logdir.setter
    def logdir(self, logdir: pathlib.Path) -> None:
        st.session_state["logdir"] = logdir

    @property
    def input_problem_name(self) -> str:
        return st.session_state["input_problem_name"]

    @input_problem_name.setter
    def input_problem_name(self, problem_name: str) -> None:
        st.session_state["input_problem_name"] = problem_name

    @property
    def is_instance_data_loaded(self) -> bool:
        return st.session_state["is_instance_data_loaded"]

    @is_instance_data_loaded.setter
    def is_instance_data_loaded(self, b: bool) -> None:
        st.session_state["is_instance_data_loaded"] = b

    @property
    def uploaded_instance_data_name(self) -> str:
        return st.session_state["uploaded_instance_data_name"]

    @uploaded_instance_data_name.setter
    def uploaded_instance_data_name(self, instance_data_name: str) -> None:
        st.session_state["uploaded_instance_data_name"] = instance_data_name

    @property
    def selected_instance_data_name(self) -> str | None:
        return st.session_state["selected_instance_data_name"]

    @selected_instance_data_name.setter
    def selected_instance_data_name(self, instance_data_name: str | None) -> None:
        st.session_state["selected_instance_data_name"] = instance_data_name

    @property
    def selected_instance_data_map(self) -> dict[str, list[str]]:
        return st.session_state["selected_instance_data_map"]

    @selected_instance_data_map.setter
    def selected_instance_data_map(self, data_map: dict[str, list[str]]) -> None:
        st.session_state["selected_instance_data_map"] = data_map

    @property
    def selected_problem_names(self) -> list[str]:
        if "checked" in self.selected_instance_data_map:
            return [
                name
                for name in self.selected_instance_data_map["expanded"]
                if len(name.split("&")) == 1
            ]
        else:
            return []

    @property
    def selected_instance_data_files(self) -> list[str]:
        if "checked" in self.selected_instance_data_map:
            return self.selected_instance_data_map["checked"]
        else:
            return []

    @property
    def selected_figure_for_instance_data(self) -> str | None:
        return st.session_state["selected_figure_for_instance_data"]

    @selected_figure_for_instance_data.setter
    def selected_figure_for_instance_data(self, fig_type: str | None) -> None:
        st.session_state["selected_figure_for_instance_data"] = fig_type

    @property
    def selected_benchmark_results(self) -> list[dict[str, str | dict[str, str | int]]]:
        return st.session_state["selected_benchmark_results"]

    @selected_benchmark_results.setter
    def selected_benchmark_results(
        self, results: dict[str, str | dict[str, str | int]]
    ) -> None:
        st.session_state["selected_benchmark_results"] = results

    @property
    def selected_benchmark_ids(self) -> list[str] | None:
        if self.selected_benchmark_results:
            return tp.cast(
                list[str],
                [result["benchmark_id"] for result in self.selected_benchmark_results],
            )

    @property
    def num_experiment_loaded(self) -> int:
        return st.session_state["num_experiment_loaded"]

    @num_experiment_loaded.setter
    def num_experiment_loaded(self, n: int) -> None:
        st.session_state["num_experiment_loaded"] = n



from __future__ import annotations

import re
from numbers import Number
from typing import Callable, Literal

import numpy as np
import pandas as pd
import plotly
import plotly.express as px

import jijbench as jb
from jijbench.exceptions.exceptions import UserFunctionFailedError
from jijbench.visualization.metrics.utils import (
    _df_has_number_array_column_target_name,
    _df_has_number_column_target_name,
    _df_has_valid_multipliers_column,
)

AXIS_LABEL_POS = Literal["top", "bottom"]


def _calc_mean_of_array(x: pd.Series, column_name: str) -> float:
    """
    Gets the mean of the elements in the specified column ,assuming this element is an np.ndarray.

    This function is intended to be used for `jb.Experiment.table`
    as experiment.table.apply(_calc_mean_of_array, axis=1, column_name=column_name)
    """
    num_occ = x["num_occurrences"]
    array = x[column_name]
    mean = np.sum(num_occ * array) / np.sum(num_occ)
    return mean


def _get_multiplier(x: pd.Series, constraint_name: str) -> float:
    """
    Gets the multiplier of given constraint name from `pd.Series`.

    This function is intended to be used for `jb.Experiment.table`
    as experiment.table.apply(_get_multiplier, axis=1, constraint_name=constraint_name)
    """
    multipliers = x["multipliers"]
    return multipliers[constraint_name]


class MetricsParallelPlot:
    def __init__(self, result: jb.Experiment) -> None:
        """Visualize the metrics of a benchmark result by parallelplot.

        Attributes:
            result (jb.Experiment): a benchmark result.
            parallelplot_axes_list (list[str]): a list of the names of all the metrics that can be plotted by `parallelplot_experiment`.
                None until `parallelplot_experiment` is called.
        """
        self.result = result

    def parallelplot_experiment(
        self,
        color_column_name: str | None = None,
        color_midpoint: float | None = None,
        additional_axes: list[str] | None = None,
        additional_axes_created_by_function: dict[str, Callable] | None = None,
        display_axes_list: list[str] | None = None,
        rename_map: dict[str, str] | None = None,
        axis_label_pos: AXIS_LABEL_POS | None = None,
        axis_label_fontsize: Number | None = None,
        title: str | None = None,
        height: Number | None = None,
        width: Number | None = None,
    ) -> plotly.graph_objects.Figure:
        """Plot the parallel plot of the experiment.

        This method creates a `plotly.graph_objects.Figure` instance and returns it.
        To view the figure, you need to call the show method of the returned instance.

        Args:
            color_column_name (str | None): the column name, and the values from this column are used to assign color to mark.
                Defaults to samplemean_total_violations or objective if those columns exist.
            color_midpoint (float | None): the midpoint of the color scale. Defaults to the mean of the color column value.
            additional_axes (list[str] | None): A list of column names for additional axes.
                The conditions for available column names are that they are elements of self.result.table.columns and that the values from the column is number.
                Defaults to None.
            additional_axes_created_by_function (dict[str, Callable]): A list of dict, where the key is the label of the axis and the value is the callable.
                The callable is applied to `self.result.table` as self.result.table.apply(callable, axis=1), and the result is added to axes.
                The callable takes a `pd.Series` and returns a number.
                Defaults to None.
            display_axes_list (list[str] | None): A list of labels for the displayed axes. This argument allows you to select and sort the axes to display.
                Check the `parallelplot_axes_list` attribute for available axes. Defaults to all axes.
            rename_map (dict[str, str] | None): A dictionary where the key is the original axis label and the value is the user-specified axis label.
                Check the original axis labels in the `parallelplot_axes_list` attribute.
                Defaults is None, the original axis labels is displayed.
            axis_label_pos (AXIS_LABEL_POS | None): the position of the axis label. Only "top" or "bottom" are accepted. Defaults to top.
            axis_label_fontsize (Number | None): the fontsize of the axis label. Defaults to None.
            title (str | None): the title of the plot. Defaults to None.
            height (Number | None): the height of the plot. Defaults to None.
            width (Number | None): the width of the plot. Defaults to None.

        Returns:
            plotly.graph_objects.Figure: the parallel plot of the experiment.

        Examples:
            The following example is the most basic usage. A parallel plot of the benchmark results is displayed.

            ```python
            from itertools import product
            import jijbench as jb
            from jijbench.visualization import MetricsParallelPlot
            import jijzept as jz

            problem = jb.get_problem("TSP")
            instance_data = jb.get_instance_data("TSP")[0][1]

            onehot_time_multipliers = [0.01, 0.1]
            onehot_location_multipliers = [0.01, 0.1]
            multipliers = [
                {"onehot_time": onehot_time_multiplier,
                "onehot_location": onehot_location_multiplier}
                for onehot_time_multiplier, onehot_location_multiplier in product(onehot_time_multipliers, onehot_location_multipliers)
            ]
            config_path = XX
            sa_sampler = jz.JijSASampler(config=config_path)
            bench = jb.Benchmark(
                params = {
                    "model": [problem],
                    "feed_dict": [instance_data],
                    "multipliers": multipliers,
                },
                solver = [sa_sampler.sample_model],
            )
            result = bench()

            mp = MetricsParallelPlot(result)
            fig = mp.parallelplot_experiment()
            fig.show()
            ```

            You can change the appearance of the graph by performing the following operations on the `plotly.graph_objects.Figure` instance returned by `parallelplot_experiment`.
            The example below changes the fontsize of the range.
            For other operations, refer to the plotly reference, https://plotly.com/python/reference/parcoords/.

            ```python
            fig.update_traces(rangefont_size=15, selector=dict(type='parcoords'))
            fig.show()
            ```

            This example gives some arguments to `parallelplot_experiment`.
            `additional_axes` argument adds the execution_time column to the plot. This is the column of result.table and its element are number.
            `additional_axes_created_by_function` argument add the values calculated from result.table to the plot.
            `rename_map` insert line breaks for long data labels to make the charts easier to read. Line breaks are done with <br>.
            `axis_label_pos` is set to bottom to avoid the data label overlapping the figure due to line breaks.

            ```python
            from itertools import product
            import numpy as np
            import pandas as pd

            import jijbench as jb
            from jijbench.visualization import MetricsParallelPlot
            import jijzept as jz
            from jijzept.sampler.openjij.sa_cpu import JijSAParameters

            problem = jb.get_problem("TSP")
            instance_data = jb.get_instance_data("TSP")[0][1]

            onehot_time_multipliers = [0.01, 0.1]
            onehot_location_multipliers = [0.01, 0.1]
            multipliers = [
                {"onehot_time": onehot_time_multiplier,
                "onehot_location": onehot_location_multiplier}
                for onehot_time_multiplier, onehot_location_multiplier in product(onehot_time_multipliers, onehot_location_multipliers)
            ]
            config_path = "XX"
            sa_parameter = JijSAParameters(num_reads=30)
            sa_sampler = jz.JijSASampler(config=config_path)
            bench = jb.Benchmark(
                params = {
                    "model": [problem],
                    "feed_dict": [instance_data],
                    "parameters": [sa_parameter],
                    "multipliers": multipliers,
                },
                solver = [sa_sampler.sample_model],
            )
            result = bench()


            def get_num_reads_from_parameters(x: pd.Series) -> float:
                return x["parameters"].num_reads

            def calc_samplemean_energy(x: pd.Series) -> float:
                num_occ	= x["num_occurrences"]
                array = x["energy"]
                mean = np.sum(num_occ * array) / np.sum(num_occ)
                return mean

            mp = MetricsParallelPlot(result)
            fig = mp.parallelplot_experiment(
                additional_axes=["execution_time"],
                additional_axes_created_by_function={
                    "num_reads": get_num_reads_from_parameters,
                    "samplemean_energy": calc_samplemean_energy,
                },
                rename_map={
                    "onehot_time_multiplier": "onehot_time<br>multiplier",
                    "onehot_location_multiplier": "onehot_location<br>multiplier",
                    "samplemean_objective": "samplemean<br>objective",
                    "samplemean_onehot_time_violations": "samplemean<br>onehot_time<br>violations",
                    "samplemean_onehot_location_violations": "samplemean<br>onehot_location<br>violations",
                    "samplemean_total_violations": "samplemean<br>total_violations",
                },
                axis_label_pos="bottom",
            )

            ```

        """
        if additional_axes is None:
            additional_axes = []
        if additional_axes_created_by_function is None:
            additional_axes_created_by_function = {}
        if axis_label_pos is None:
            axis_label_pos = "top"
        if not (axis_label_pos in ["top", "bottom"]):
            raise ValueError(
                f"axis_label_pos must be 'top' or 'bottom', but {axis_label_pos} is given."
            )

        result_table = self.result.table

        # The key is a column name (str), and the value is the data of each column (pd.Series).
        data_to_create_df_parallelplot = {}

        # Displayed data specified by user
        for display_column in additional_axes:
            # Check if the column exists and the elements is number.
            if not _df_has_number_column_target_name(result_table, display_column):
                raise TypeError(
                    f"{display_column}is not a column with number elements."
                )
            data_to_create_df_parallelplot[display_column] = result_table[
                display_column
            ]

        # Data generated by user custom functions
        for column_name, func in additional_axes_created_by_function.items():
            # TODO: デコレータで書き直した方が可読性が上がると思われる。エラーレイズと返り値がnumberであることのチェックをデコレータ内で行う
            try:
                data_to_create_df_parallelplot[column_name] = result_table.apply(
                    func, axis=1
                )
            except Exception as e:
                msg = f'An error occurred inside your function. Please check "{func.__name__}" in additional_axes_created_by_function. -> {e}'
                raise UserFunctionFailedError(msg)
            for value in data_to_create_df_parallelplot[column_name].values:
                if not isinstance(value, Number):
                    raise TypeError(
                        f"{column_name} is not a column with number elements."
                    )

        # multiplires (If self.result has a valid multipliers column)
        if _df_has_valid_multipliers_column(result_table):
            data_to_create_df_parallelplot.update(
                result_table.filter(regex="multipliers").to_dict(orient="list")
            )

        # num_feasible
        if _df_has_number_column_target_name(result_table, "num_feasible"):
            data_to_create_df_parallelplot["num_feasible"] = result_table[
                "num_feasible"
            ]

        # objective
        if _df_has_number_array_column_target_name(result_table, "objective"):
            data_to_create_df_parallelplot["samplemean_objective"] = result_table.apply(
                _calc_mean_of_array, axis=1, column_name="objective"
            )

        # violations
        for violation_column_name in result_table.columns[
            result_table.columns.str.contains("violations")
        ]:
            if _df_has_number_array_column_target_name(
                result_table, violation_column_name
            ):
                data_to_create_df_parallelplot[
                    "samplemean_" + violation_column_name
                ] = result_table.apply(
                    _calc_mean_of_array,
                    axis=1,
                    column_name=violation_column_name,
                )

        self.df_parallelplot = df_parallelplot = pd.DataFrame(
            data_to_create_df_parallelplot
        )

        if display_axes_list is None:
            display_axes_list = self.parallelplot_axes_list
        self.df_parallelplot_displayed = df_parallelplot_displayed = df_parallelplot[
            display_axes_list
        ]

        if color_column_name is None:
            if "samplemean_total_violations" in df_parallelplot_displayed.columns:
                color_column_name = "samplemean_total_violations"
            elif "samplemean_objective" in df_parallelplot_displayed.columns:
                color_column_name = "samplemean_objective"

        if color_midpoint is None and color_column_name is not None:
            color_midpoint = df_parallelplot_displayed[color_column_name].mean()

        fig = px.parallel_coordinates(
            df_parallelplot_displayed.reset_index(drop=True),
            color=color_column_name,
            labels=rename_map,
            color_continuous_scale=px.colors.diverging.Tealrose,
            color_continuous_midpoint=color_midpoint,
            title=title,
            height=height,
            width=width,
        )
        fig.update_traces(labelside=axis_label_pos, selector=dict(type="parcoords"))
        fig.update_traces(
            labelfont_size=axis_label_fontsize, selector=dict(type="parcoords")
        )

        return fig

    @property
    def parallelplot_axes_list(self) -> list[str]:
        if hasattr(self, "df_parallelplot"):
            return list(self.df_parallelplot.columns)
        else:
            return []
